from concurrent.futures import Future
import logging
from pathlib import Path
import re
import shutil
import subprocess
import sys
import time
from typing import (
    TYPE_CHECKING,
    Dict,
    List,
    Optional,
    Tuple,
    Union,
)
import uuid

from safety_schemas.models.events import Event, EventType
from safety_schemas.models.events.types import ToolType
from safety_schemas.models.events.payloads import (
    CodebaseDetectionStatusPayload,
    CodebaseSetupCompletedPayload,
    CodebaseSetupResponseCreatedPayload,
    DependencyFile,
    FirewallConfiguredPayload,
    FirewallDisabledPayload,
    FirewallSetupCompletedPayload,
    FirewallSetupResponseCreatedPayload,
    InitExitStep,
    InitExitedPayload,
    InitScanCompletedPayload,
    PackageInstalledPayload,
    PackageUninstalledPayload,
    PackageUpdatedPayload,
    CommandExecutedPayload,
    ToolCommandExecutedPayload,
    CommandErrorPayload,
    AliasConfig,
    IndexConfig,
    ToolStatus,
    CommandParam,
    ProcessStatus,
    FirewallHeartbeatPayload,
    InitStartedPayload,
    AuthStartedPayload,
    AuthCompletedPayload,
)
import typer

from ..event_bus import EventBus
from ..types.base import InternalEventType, InternalPayload

from .creation import (
    create_event,
)
from .data import (
    clean_parameter,
    get_command_path,
    get_root_context,
    scrub_sensitive_value,
    translate_param_source,
)
from .conditions import conditional_emitter, should_emit_firewall_heartbeat


if TYPE_CHECKING:
    from safety.models import SafetyCLI, ToolResult
    from safety.cli_util import CustomContext
    from safety.init.types import FirewallConfigStatus

logger = logging.getLogger(__name__)


@conditional_emitter
def send_and_flush(event_bus: "EventBus", event: Event) -> Optional[Future]:
    """
    Emit an event and immediately flush the event bus without closing it.

    Args:
        event_bus: The event bus to emit on
        event: The event to emit
    """
    future = event_bus.emit(event)

    # Create and emit flush event
    flush_payload = InternalPayload()
    flush_event = create_event(
        payload=flush_payload, event_type=InternalEventType.FLUSH_SECURITY_TRACES
    )

    # Emit flush event and wait for it to complete
    flush_future = event_bus.emit(flush_event)

    # Wait for both events to complete
    if future:
        try:
            future.result(timeout=0.5)
        except Exception:
            logger.error("Emit Failed %s (%s)", event.type, event.id)

    if flush_future:
        try:
            return flush_future.result(timeout=0.5)
        except Exception:
            logger.error("Flush Failed for event %s", event.id)

    return None


@conditional_emitter(conditions=[should_emit_firewall_heartbeat])
def emit_firewall_heartbeat(
    event_bus: "EventBus", ctx: Optional["CustomContext"], *, tools: List[ToolStatus]
):
    payload = FirewallHeartbeatPayload(tools=tools)
    event = create_event(payload=payload, event_type=EventType.FIREWALL_HEARTBEAT)

    event_bus.emit(event)


@conditional_emitter
def emit_firewall_disabled(
    event_bus: "EventBus",
    ctx: Optional["CustomContext"] = None,
    *,
    reason: Optional[str],
):
    payload = FirewallDisabledPayload(reason=reason)
    event = create_event(payload=payload, event_type=EventType.FIREWALL_DISABLED)

    event_bus.emit(event)


def status_to_tool_status(status: "FirewallConfigStatus") -> List[ToolStatus]:
    tools = []
    for tool_type, configs in status.items():
        alias_config = (
            configs["alias"] if isinstance(configs["alias"], AliasConfig) else None
        )
        index_config = (
            configs["index"] if isinstance(configs["index"], IndexConfig) else None
        )

        tool = tool_type.value
        command_path = shutil.which(tool)
        reachable = False
        version = "unknown"

        if command_path:
            args = [command_path, "--version"]
            result = subprocess.run(args, capture_output=True, text=True)

            if result.returncode == 0:
                output = result.stdout
                reachable = True

                # Extract version
                version_match = re.search(r"(\d+\.\d+(?:\.\d+)?)", output)
                if version_match:
                    version = version_match.group(1)
        else:
            command_path = tool

        tool = ToolStatus(
            type=tool_type,
            command_path=command_path,
            version=version,
            reachable=reachable,
            alias_config=alias_config,
            index_config=index_config,
        )
        tools.append(tool)

    return tools


@conditional_emitter
def emit_firewall_configured(
    event_bus: "EventBus",
    ctx: Optional["CustomContext"] = None,
    *,
    status: "FirewallConfigStatus",
):
    tools = status_to_tool_status(status)

    payload = FirewallConfiguredPayload(tools=tools)

    event = create_event(payload=payload, event_type=EventType.FIREWALL_CONFIGURED)

    event_bus.emit(event)


@conditional_emitter
def emit_diff_operations(
    event_bus: "EventBus",
    ctx: "CustomContext",
    *,
    added: Dict[str, str],
    removed: Dict[str, str],
    updated: Dict[str, Tuple[str, str]],
    by_tool: ToolType,
):
    obj: "SafetyCLI" = ctx.obj
    correlation_id = obj.correlation_id

    if (added or removed or updated) and not correlation_id:
        correlation_id = obj.correlation_id = str(uuid.uuid4())

    def emit_package_event(event_bus, correlation_id, payload, event_type):
        event = create_event(
            payload=payload,
            event_type=event_type,
            correlation_id=correlation_id,
        )
        event_bus.emit(event)

    for package_name, version in added.items():
        emit_package_event(
            event_bus,
            correlation_id,
            PackageInstalledPayload(
                package_name=package_name, version=version, tool=by_tool
            ),
            EventType.PACKAGE_INSTALLED,
        )

    for package_name, version in removed.items():
        emit_package_event(
            event_bus,
            correlation_id,
            PackageUninstalledPayload(
                package_name=package_name, version=version, tool=by_tool
            ),
            EventType.PACKAGE_UNINSTALLED,
        )

    for package_name, (previous_version, current_version) in updated.items():
        emit_package_event(
            event_bus,
            correlation_id,
            PackageUpdatedPayload(
                package_name=package_name,
                previous_version=previous_version,
                current_version=current_version,
                tool=by_tool,
            ),
            EventType.PACKAGE_UPDATED,
        )


@conditional_emitter
def emit_tool_command_executed(
    event_bus: "EventBus", ctx: "CustomContext", *, tool: ToolType, result: "ToolResult"
) -> None:
    correlation_id = ctx.obj.correlation_id

    if not correlation_id:
        correlation_id = ctx.obj.correlation_id = str(uuid.uuid4())

    process = result.process

    payload = ToolCommandExecutedPayload(
        tool=tool,
        raw_command=[clean_parameter("", arg) for arg in process.args],
        duration_ms=result.duration_ms,
        status=ProcessStatus(
            stdout=process.stdout, stderr=process.stderr, return_code=process.returncode
        ),
    )

    # Scrub after binary coercion to str
    if payload.status.stdout:
        payload.status.stdout = scrub_sensitive_value(payload.status.stdout)
    if payload.status.stderr:
        payload.status.stderr = scrub_sensitive_value(payload.status.stderr)

    event = create_event(
        correlation_id=correlation_id,
        payload=payload,
        event_type=EventType.TOOL_COMMAND_EXECUTED,
    )

    event_bus.emit(event)


@conditional_emitter
def emit_command_executed(
    event_bus: "EventBus", ctx: "CustomContext", *, returned_code: int
) -> None:
    root_context = get_root_context(ctx)
    NA = ""

    started_at = getattr(root_context, "started_at", None) if root_context else None
    if started_at is not None:
        duration_ms = int((time.monotonic() - started_at) * 1000)
    else:
        duration_ms = 1

    command_name = ctx.command.name if ctx.command.name is not None else NA
    raw_command = [clean_parameter("", arg) for arg in sys.argv]

    params: List[CommandParam] = []

    for idx, param in enumerate(ctx.command.params):
        param_name = param.name if param.name is not None else NA
        param_value = ctx.params.get(param_name)

        # Scrub the parameter value if sensitive
        scrubbed_value = clean_parameter(param_name, param_value)

        # Determine parameter source using Click's API
        click_source = ctx.get_parameter_source(param_name)
        source = translate_param_source(click_source)

        display_name = param_name if param_name else None

        params.append(
            CommandParam(
                position=idx, name=display_name, value=scrubbed_value, source=source
            )
        )

    payload = CommandExecutedPayload(
        command_name=command_name,
        command_path=get_command_path(ctx),
        raw_command=raw_command,
        parameters=params,
        duration_ms=duration_ms,
        status=ProcessStatus(
            return_code=returned_code,
        ),
    )

    event = create_event(
        correlation_id=ctx.obj.correlation_id,
        payload=payload,
        event_type=EventType.COMMAND_EXECUTED,
    )

    try:
        if future := event_bus.emit(event):
            future.result(timeout=0.5)
    except Exception:
        logger.error("Emit Failed %s (%s)", event.type, event.id)


@conditional_emitter
def emit_command_error(
    event_bus: "EventBus",
    ctx: "CustomContext",
    *,
    message: str,
    traceback: Optional[str] = None,
) -> None:
    """
    Emit a CommandErrorEvent with sensitive data scrubbed.
    """
    # Get command name from context if available
    command_name = getattr(ctx, "command", None)
    if command_name and command_name.name:
        command_name = command_name.name

    scrub_traceback = None
    if traceback:
        scrub_traceback = scrub_sensitive_value(traceback)

    command_path = get_command_path(ctx)
    raw_command = [scrub_sensitive_value(arg) for arg in sys.argv]

    payload = CommandErrorPayload(
        command_name=command_name,
        raw_command=raw_command,
        command_path=command_path,
        error_message=scrub_sensitive_value(message),
        stacktrace=scrub_traceback,
    )

    event = create_event(
        payload=payload,
        event_type=EventType.COMMAND_ERROR,
    )

    event_bus.emit(event)


def emit_init_started(
    event_bus: "EventBus", ctx: Union["CustomContext", typer.Context]
) -> None:
    """
    Emit an InitStartedEvent and store it as a pending event in SafetyCLI object.

    Args:
        event_bus: The event bus to emit on
        ctx: The Click context containing the SafetyCLI object
    """
    obj: "SafetyCLI" = ctx.obj

    if not obj.correlation_id:
        obj.correlation_id = str(uuid.uuid4())

    payload = InitStartedPayload()
    event = create_event(
        correlation_id=obj.correlation_id,
        payload=payload,
        event_type=EventType.INIT_STARTED,
    )

    if not send_and_flush(event_bus, event):
        # Store as pending event
        obj.pending_events.append(event)


def emit_auth_started(event_bus: "EventBus", ctx: "CustomContext") -> None:
    """
    Emit an AuthStartedEvent and store it as a pending event in SafetyCLI object.

    Args:
        event_bus: The event bus to emit on
        ctx: The Click context containing the SafetyCLI object
    """
    obj: "SafetyCLI" = ctx.obj

    if not obj.correlation_id:
        obj.correlation_id = str(uuid.uuid4())

    payload = AuthStartedPayload()
    event = create_event(
        correlation_id=obj.correlation_id,
        payload=payload,
        event_type=EventType.AUTH_STARTED,
    )

    if not send_and_flush(event_bus, event):
        # Store as pending event
        obj.pending_events.append(event)


@conditional_emitter
def emit_auth_completed(
    event_bus: "EventBus",
    ctx: "CustomContext",
    *,
    success: bool = True,
    error_message: Optional[str] = None,
) -> None:
    """
    Emit an AuthCompletedEvent and submit all pending events together.

    Args:
        event_bus: The event bus to emit on
        ctx: The Click context containing the SafetyCLI object
        success: Whether authentication was successful
        error_message: Optional error message if authentication failed
    """
    obj: "SafetyCLI" = ctx.obj

    if not obj.correlation_id:
        obj.correlation_id = str(uuid.uuid4())

    payload = AuthCompletedPayload(success=success, error_message=error_message)

    event = create_event(
        correlation_id=obj.correlation_id,
        payload=payload,
        event_type=EventType.AUTH_COMPLETED,
    )

    for pending_event in obj.pending_events:
        event_bus.emit(pending_event)

    obj.pending_events.clear()

    # Emit auth completed event and flush
    send_and_flush(event_bus, event)


@conditional_emitter
def emit_firewall_setup_response_created(
    event_bus: "EventBus",
    ctx: Union["CustomContext", typer.Context],
    *,
    user_consent_requested: bool,
    user_consent: Optional[bool] = None,
) -> None:
    obj: "SafetyCLI" = ctx.obj

    if not obj.correlation_id:
        obj.correlation_id = str(uuid.uuid4())

    payload = FirewallSetupResponseCreatedPayload(
        user_consent_requested=user_consent_requested, user_consent=user_consent
    )

    event = create_event(
        correlation_id=obj.correlation_id,
        payload=payload,
        event_type=EventType.FIREWALL_SETUP_RESPONSE_CREATED,
    )

    # Emit and flush
    send_and_flush(event_bus, event)


@conditional_emitter
def emit_codebase_setup_response_created(
    event_bus: "EventBus",
    ctx: Union["CustomContext", typer.Context],
    *,
    user_consent_requested: bool,
    user_consent: Optional[bool] = None,
) -> None:
    obj: "SafetyCLI" = ctx.obj

    if not obj.correlation_id:
        obj.correlation_id = str(uuid.uuid4())

    payload = CodebaseSetupResponseCreatedPayload(
        user_consent_requested=user_consent_requested, user_consent=user_consent
    )

    event = create_event(
        correlation_id=obj.correlation_id,
        payload=payload,
        event_type=EventType.CODEBASE_SETUP_RESPONSE_CREATED,
    )

    # Emit and flush
    send_and_flush(event_bus, event)


@conditional_emitter
def emit_codebase_detection_status(
    event_bus: "EventBus",
    ctx: Union["CustomContext", typer.Context],
    *,
    detected: bool,
    detected_files: Optional[List[Path]] = None,
) -> None:
    obj: "SafetyCLI" = ctx.obj

    if not obj.correlation_id:
        obj.correlation_id = str(uuid.uuid4())

    payload = CodebaseDetectionStatusPayload(
        detected=detected,
        dependency_files=[
            DependencyFile(file_path=str(file)) for file in detected_files
        ]
        if detected_files
        else None,
    )

    event = create_event(
        correlation_id=obj.correlation_id,
        payload=payload,
        event_type=EventType.CODEBASE_DETECTION_STATUS,
    )

    # Emit and flush
    send_and_flush(event_bus, event)


@conditional_emitter
def emit_init_scan_completed(
    event_bus: "EventBus",
    ctx: Union["CustomContext", typer.Context],
    *,
    scan_id: Optional[str],
) -> None:
    obj: "SafetyCLI" = ctx.obj

    if not obj.correlation_id:
        obj.correlation_id = str(uuid.uuid4())

    payload = InitScanCompletedPayload(scan_id=scan_id)

    event = create_event(
        correlation_id=obj.correlation_id,
        payload=payload,
        event_type=EventType.INIT_SCAN_COMPLETED,
    )

    # Emit and flush
    send_and_flush(event_bus, event)


@conditional_emitter
def emit_codebase_setup_completed(
    event_bus: "EventBus",
    ctx: Union["CustomContext", typer.Context],
    *,
    is_created: bool,
    codebase_id: Optional[str] = None,
) -> None:
    obj: "SafetyCLI" = ctx.obj

    if not obj.correlation_id:
        obj.correlation_id = str(uuid.uuid4())

    payload = CodebaseSetupCompletedPayload(
        is_created=is_created, codebase_id=codebase_id
    )

    event = create_event(
        correlation_id=obj.correlation_id,
        payload=payload,
        event_type=EventType.CODEBASE_SETUP_COMPLETED,
    )

    # Emit and flush
    send_and_flush(event_bus, event)


@conditional_emitter
def emit_firewall_setup_completed(
    event_bus: "EventBus",
    ctx: "CustomContext",
    *,
    status: "FirewallConfigStatus",
) -> None:
    obj: "SafetyCLI" = ctx.obj

    if not obj.correlation_id:
        obj.correlation_id = str(uuid.uuid4())

    tools = status_to_tool_status(status)

    payload = FirewallSetupCompletedPayload(
        tools=tools,
    )

    event = create_event(
        correlation_id=obj.correlation_id,
        payload=payload,
        event_type=EventType.FIREWALL_SETUP_COMPLETED,
    )

    # Emit and flush
    send_and_flush(event_bus, event)


@conditional_emitter
def emit_init_exited(
    event_bus: "EventBus",
    ctx: Union["CustomContext", typer.Context],
    *,
    exit_step: InitExitStep,
) -> None:
    obj: "SafetyCLI" = ctx.obj

    if not obj.correlation_id:
        obj.correlation_id = str(uuid.uuid4())

    payload = InitExitedPayload(exit_step=exit_step)

    event = create_event(
        correlation_id=obj.correlation_id,
        payload=payload,
        event_type=EventType.INIT_EXITED,
    )

    # Emit and flush
    send_and_flush(event_bus, event)



from collections import namedtuple
from typing import List, Dict, Any, Optional, Tuple

from safety.formatter import FormatterAPI
from safety.util import get_basic_announcements


class BareReport(FormatterAPI):
    """
    Bare report, for command line tools.
    """

    def render_vulnerabilities(self, announcements: List[Dict[str, Any]], vulnerabilities: List[Any],
                               remediations: Any, full: bool, packages: List[Any], fixes: Tuple = ()) -> str:
        """
        Renders vulnerabilities in a bare format.

        Args:
            announcements (List[Dict[str, Any]]): List of announcements.
            vulnerabilities (List[Any]): List of vulnerabilities.
            remediations (Any): Remediation data.
            full (bool): Flag indicating full output.
            packages (List[Any]): List of packages.
            fixes (Tuple, optional): Tuple of fixes.

        Returns:
            str: Rendered vulnerabilities.
        """
        parsed_announcements = []
        Announcement = namedtuple("Announcement", ["name"])

        for announcement in get_basic_announcements(announcements, include_local=False):
            normalized_message = "-".join(announcement.get('message', 'none').lower().split())
            parsed_announcements.append(Announcement(name=normalized_message))

        announcements_to_render = [announcement.name for announcement in parsed_announcements]
        affected_packages = list(set([v.package_name for v in vulnerabilities if not v.ignored]))

        return " ".join(announcements_to_render + affected_packages)

    def render_licenses(self, announcements: List[Dict[str, Any]], packages_licenses: List[Dict[str, Any]]) -> str:
        """
        Renders licenses in a bare format.

        Args:
            announcements (List[Dict[str, Any]]): List of announcements.
            packages_licenses (List[Dict[str, Any]]): List of package licenses.

        Returns:
            str: Rendered licenses.
        """
        parsed_announcements = []

        for announcement in get_basic_announcements(announcements):
            normalized_message = "-".join(announcement.get('message', 'none').lower().split())
            parsed_announcements.append({'license': normalized_message})

        announcements_to_render = [announcement.get('license') for announcement in parsed_announcements]

        licenses = list(set([pkg_li.get('license') for pkg_li in packages_licenses]))
        sorted_licenses = sorted(licenses)
        return " ".join(announcements_to_render + sorted_licenses)

    def render_announcements(self, announcements: List[Dict[str, Any]]) -> None:
        """
        Renders announcements in a bare format.

        Args:
            announcements (List[Dict[str, Any]]): List of announcements.
        """
        print('render_announcements bare')



"""
Typosquatting detection for various tools.
"""

import logging
import nltk
from typing import Tuple, List

from safety.console import main_console as console
from rich.prompt import Prompt

logger = logging.getLogger(__name__)


class TyposquattingProtection:
    """
    Base class for typosquatting detection.
    """

    def __init__(self, popular_packages: List[str]):
        self.popular_packages = popular_packages

    def check_package(self, package_name: str) -> Tuple[bool, str]:
        """
        Check if a package name is likely to be a typosquatting attempt.

        Args:
            package_name: Name of the package to check

        Returns:
            Tuple of (is_valid, suggested_package_name)
        """
        max_edit_distance = 2 if len(package_name) > 5 else 1

        if package_name in self.popular_packages:
            return (True, package_name)

        for pkg in self.popular_packages:
            if (
                abs(len(pkg) - len(package_name)) <= max_edit_distance
                and nltk.edit_distance(pkg, package_name) <= max_edit_distance
            ):
                return (False, pkg)

        return (True, package_name)

    def coerce(self, dependency_name: str) -> str:
        """
        Coerce a package name to its correct name if it is a typosquatting attempt.

        Args:
            dependency_name: Name of the package to coerce

        Returns:
            str: Coerced package name
        """
        (valid, candidate_package_name) = self.check_package(dependency_name)

        if not valid:
            prompt = f"You are about to install {dependency_name} package. Did you mean to install {candidate_package_name}?"
            answer = Prompt.ask(
                prompt=prompt,
                choices=["y", "n"],
                default="y",
                show_default=True,
                console=console,
            ).lower()
            if answer == "y":
                return candidate_package_name

        return dependency_name



# type: ignore
from sys import platform
import unittest
from unittest.mock import patch
from pathlib import Path
import tempfile
import shutil

from datetime import datetime, timezone

import pytest

from safety.tool.interceptors.unix import UnixAliasInterceptor


@pytest.mark.unix_only
@pytest.mark.skipif(
    platform not in ["linux", "linux2", "darwin"], reason="Unix-specific tests"
)
class TestUnixAliasInterceptor(unittest.TestCase):
    def setUp(self):
        self.temp_dir = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, self.temp_dir)

    @patch("safety.tool.interceptors.unix.Path.home")
    @patch("safety.tool.interceptors.base.datetime")
    @patch("safety.tool.interceptors.base.get_version")
    def test_interceptors_all_tools(self, mock_version, mock_datetime, mock_home):
        mock_home.return_value = Path(self.temp_dir)
        mock_version.return_value = "1.0.0"
        mock_now = datetime(2024, 1, 1, tzinfo=timezone.utc)
        mock_datetime.now.return_value = mock_now

        safety_config_user_dir = Path(self.temp_dir) / ".safety"

        with patch(
            "safety.tool.interceptors.unix.USER_CONFIG_DIR", safety_config_user_dir
        ):
            interceptor = UnixAliasInterceptor()
            result = interceptor.install_interceptors()

            self.assertTrue(result)

            rc_files = (".zshrc", ".bashrc", ".bash_profile")

            rc_paths = [Path(self.temp_dir) / rc for rc in rc_files]
            safety_profile_path = Path(self.temp_dir) / ".safety" / ".safety_profile"

            self.assertTrue(
                all(path.exists() for path in rc_paths), "Not all rc_paths exist"
            )
            self.assertTrue(safety_profile_path.exists())

            # test the content of the generated files
            expected_rc_content = (
                "# >>> Safety >>>\n"
                f'[ -f "{safety_profile_path}" ] && . "{safety_profile_path}"\n'
                "# <<< Safety <<<\n"
            )

            expected_safety_profile_content = (
                "# >>> Safety >>>\n"
                "# DO NOT EDIT THIS FILE DIRECTLY\n"
                f"# Last updated at: {mock_now.isoformat()}\n"
                "# Updated by: safety v1.0.0\n"
                'alias pip="safety pip"\n'
                'alias pip3="safety pip3"\n'
                'alias pip3.8="safety pip3.8"\n'
                'alias pip3.9="safety pip3.9"\n'
                'alias pip3.10="safety pip3.10"\n'
                'alias pip3.11="safety pip3.11"\n'
                'alias pip3.12="safety pip3.12"\n'
                'alias pip3.13="safety pip3.13"\n'
                'alias pip3.14="safety pip3.14"\n'
                'alias poetry="safety poetry"\n'
                'alias uv="safety uv"\n'
                "# <<< Safety <<<\n"
            )

            for rc_path in rc_paths:
                self.assertEqual(rc_path.read_text(), expected_rc_content)

            self.assertEqual(
                safety_profile_path.read_text(), expected_safety_profile_content
            )

            # Let's test remove_interceptors
            result = interceptor.remove_interceptors()

            self.assertTrue(result)
            self.assertTrue(
                all(path.exists() for path in rc_paths), "Some rc_paths were removed"
            )

            for rc_path in rc_paths:
                self.assertEqual(rc_path.read_text(), "")

            self.assertFalse(safety_profile_path.exists())

    def test_install_interceptors_nonexistent_tool(self):
        interceptor = UnixAliasInterceptor()
        result = interceptor.install_interceptors(["nonexistent"])
        self.assertFalse(result)

    def test_uninstall_interceptors_all_tools(self):
        interceptor = UnixAliasInterceptor()
        result = interceptor.install_interceptors()
        self.assertTrue(result)



import base64
import json

import typer


def index_credentials(ctx: typer.Context) -> str:
    """
    Returns the index credentials for the current context.
    This should be used together with user:index_credential for index
    basic auth.

    Args:
        ctx (typer.Context): The context.

    Returns:
        str: The index credentials.
    """
    api_key = None
    token = None

    if auth := getattr(ctx.obj, "auth", None):
        client = auth.client
        token = client.token.get("access_token") if client.token else None
        api_key = client.api_key

    auth_envelop = json.dumps(
        {
            "version": "1.0",
            "access_token": token,
            "api_key": api_key,
            "project_id": ctx.obj.project.id if ctx.obj.project else None,
        }
    )
    return base64.urlsafe_b64encode(auth_envelop.encode("utf-8")).decode("utf-8")



from typing import List

import typer
from safety.tool.intents import ToolIntentionType
from safety.tool.pip.parser import PipParser
from safety.tool.auth import index_credentials
from ..pip.command import PipCommand, PipInstallCommand, PipGenericCommand
from safety_schemas.models.events.types import ToolType

UV_LOCK = "safety-uv.lock"


class UvCommand(PipCommand):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._name = ["uv"]

    def get_tool_type(self) -> ToolType:
        return ToolType.UV

    def get_lock_path(self) -> str:
        return UV_LOCK

    def get_package_list_command(self) -> List[str]:
        return [*self._name, "pip", "list", "--format=json"]

    def should_track_state(self) -> bool:
        should_track = super().should_track_state()

        if should_track:
            return True

        command_str = " ".join(self._args).lower()

        package_modifying_commands = [
            "sync",
        ]

        return any(cmd in command_str for cmd in package_modifying_commands)

    def env(self, ctx: typer.Context) -> dict:
        env = super().env(ctx)

        env.update(
            {
                "UV_INDEX_SAFETY_USERNAME": "user",
                "UV_INDEX_SAFETY_PASSWORD": index_credentials(ctx),
            }
        )

        return env

    @classmethod
    def from_args(cls, args: List[str], **kwargs):
        pip_parser = PipParser()
        is_pip_interface = args and args[0] == "pip"

        to_parse = args[1:] if is_pip_interface else args

        if intention := pip_parser.parse(to_parse):
            if intention.intention_type is ToolIntentionType.ADD_PACKAGE:
                return UvInstallCommand(to_parse, intention=intention, **kwargs)

        # No an install but still a pip interface command
        if is_pip_interface:
            to_parse = args

        return UvGenericCommand(to_parse, **kwargs)


class UvInstallCommand(PipInstallCommand, UvCommand):
    def get_command_name(self) -> List[str]:
        return ["uv", "pip"]


class UvGenericCommand(PipGenericCommand, UvCommand):
    def get_command_name(self) -> List[str]:
        return ["uv"]



import logging
from enum import Enum
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Generator,
    List,
    Optional,
    Tuple,
    Union,
    Literal,
)
from pydantic import BaseModel
import typer

from safety.auth.constants import SAFETY_PLATFORM_URL
from safety.errors import SafetyException
from safety.scan.command import (
    ScannableEcosystems,
    initialize_file_finder,
    scan_project_directory,
)
from safety.scan.main import (
    download_policy,
    load_policy_file,
    process_files,
    resolve_policy,
)
from safety_schemas.models import (
    Ecosystem,
    FileModel,
    FileType,
    RemediationModel,
    ReportModel,
    ReportSchemaVersion,
    ScanType,
    VulnerabilitySeverityLabels,
    MetadataModel,
    TelemetryModel,
    ProjectModel,
    Stage,
    AuthenticationType,
)

from safety.scan.util import GIT
from safety.util import build_telemetry_data


# Define typed models for scan results
class ScanResultType(str, Enum):
    """Types of scan results that can be yielded by the init_scan function"""

    INIT = "init"
    PROGRESS = "progress"
    UPLOADING = "uploading"
    STATUS = "status"
    COMPLETE = "complete"


class BaseScanResult(BaseModel):
    """Base class for all scan results with common attributes"""

    # No fields here - each subclass will define its own type
    pass


class InitScanResult(BaseScanResult):
    """Initial scan result with basic dependency info"""

    type: Literal[ScanResultType.INIT]
    dependencies: int
    progress: int = 0

    class Config:
        frozen = True  # Make immutable


class ProgressScanResult(BaseScanResult):
    """Progress update during scanning with current counts"""

    type: Literal[ScanResultType.PROGRESS]
    percent: int
    dependencies: int
    critical: Optional[int] = None
    high: Optional[int] = None
    medium: Optional[int] = None
    low: Optional[int] = None
    others: Optional[int] = None
    fixes: Optional[int] = None
    fixed_vulns: Optional[int] = None
    file: str
    file_pkg_count: int
    file_count: int
    venv_count: int
    vulns_count: int

    class Config:
        frozen = True  # Make immutable


class CompleteScanResult(BaseScanResult):
    """Final scan result with complete vulnerability counts"""

    type: Literal[ScanResultType.COMPLETE]
    scan_id: Optional[str] = None
    percent: int = 100
    dependencies: int
    critical: int
    high: int
    medium: int
    low: int
    others: int
    vulns_count: int
    fixes: int
    fixed_vulns: int
    codebase_url: Optional[str] = None

    class Config:
        frozen = True  # Make immutable


class StatusScanResult(BaseScanResult):
    """Generic status update that can be used for any process"""

    type: Literal[ScanResultType.STATUS]
    message: str
    action: str  # The specific action being performed (e.g., "analyzing", "preparing")
    percent: Optional[int] = None

    class Config:
        frozen = True  # Make immutable


class UploadingScanResult(BaseScanResult):
    """Status update when uploading results to server"""

    type: Literal[ScanResultType.UPLOADING]
    message: str
    percent: Optional[int] = None

    class Config:
        frozen = True  # Make immutable


# Union type for all possible result types
ScanResult = Union[
    InitScanResult,
    ProgressScanResult,
    StatusScanResult,
    UploadingScanResult,
    CompleteScanResult,
]

LOG = logging.getLogger(__name__)

if TYPE_CHECKING:
    from safety_schemas.models import (
        ConfigModel,
        ProjectModel,
        MetadataModel,
        TelemetryModel,
        ReportModel,
        FileModel,
    )


def init_scan(
    ctx: Any,
    target: Path,
    config: "ConfigModel",
    metadata: "MetadataModel",
    telemetry: "TelemetryModel",
    project: "ProjectModel",
    use_server_matching: bool = False,
) -> Generator[ScanResult, None, Tuple["ReportModel", List["FileModel"]]]:
    """
    Core scanning logic that yields results as they become available.
    Contains no UI-related code - purely logic for scanning.

    Args:
        ctx: The context object with necessary configurations
        target: The target directory to scan
        config: The application configuration
        metadata: Metadata to include in the report
        telemetry: Telemetry data to include in the report
        project: The project object
        version: The schema version
        use_server_matching: Whether to use server-side vulnerability matching

    Yields:
        Dict containing scan progress information and results as they become available

    Returns:
        Tuple containing the final report model and list of files
    """
    # Emit status that scan is starting
    yield StatusScanResult(
        type=ScanResultType.STATUS,
        message="Starting safety scan",
        action="initializing",
        percent=0,
    )

    # Initialize ecosystems
    ecosystems = [Ecosystem(member.value) for member in list(ScannableEcosystems)]

    # Initialize file finder and locate project files
    from rich.console import Console

    console = Console()
    console.quiet = True

    yield StatusScanResult(
        type=ScanResultType.STATUS,
        message="Locating project files",
        action="discovering",
        percent=5,
    )

    file_finder = initialize_file_finder(ctx, target, None, ecosystems)

    yield StatusScanResult(
        type=ScanResultType.STATUS,
        message="Scanning project directory",
        action="scanning",
        percent=10,
    )

    _, file_paths = scan_project_directory(file_finder, console)

    total_files = sum(len(file_set) for file_set in file_paths.values())

    yield StatusScanResult(
        type=ScanResultType.STATUS,
        message=f"Found {total_files} files to analyze",
        action="analyzing",
        percent=15,
    )

    # Initialize counters and data structures
    files: List[FileModel] = []
    count = 0  # Total dependencies processed
    affected_count = 0
    critical_vulns_count = 0
    high_vulns_count = 0
    medium_vulns_count = 0
    low_vulns_count = 0
    others_vulns_count = 0
    vulns_count = 0
    fixes_count = 0
    total_resolved_vulns = 0
    file_count = 0
    venv_count = 0
    scan_id = None
    # Count the total number of files across all types

    # Initial yield with dependency info
    yield InitScanResult(type=ScanResultType.INIT, dependencies=count)

    # Status update before processing files
    yield StatusScanResult(
        type=ScanResultType.STATUS,
        message="Processing files for dependencies and vulnerabilities",
        action="analyzing",
        percent=20,
    )

    # Process each file for dependencies and vulnerabilities
    for idx, (path, analyzed_file) in enumerate(
        process_files(
            paths=file_paths,
            config=config,
            use_server_matching=use_server_matching,
            obj=ctx.obj,
            target=target,
        )
    ):
        # Calculate progress percentage
        # Calculate progress and ensure it never exceeds 100%
        if total_files > 0:
            progress = min(int((idx + 1) / total_files * 100), 100)
        else:
            progress = 100

        # Update counts for dependencies
        file_pkg_count = len(analyzed_file.dependency_results.dependencies)
        count += file_pkg_count

        # Track environment/file types
        if analyzed_file.file_type is FileType.VIRTUAL_ENVIRONMENT:
            venv_count += 1
        else:
            file_count += 1

        # Get affected specifications
        affected_specifications = (
            analyzed_file.dependency_results.get_affected_specifications()
        )
        affected_count += len(affected_specifications)

        # Count vulnerabilities by severity
        current_critical = 0
        current_high = 0
        current_medium = 0
        current_low = 0
        current_others = 0
        current_fixes = 0
        current_resolved_vulns = 0

        # Process each affected specification
        for spec in affected_specifications:
            # Access vulnerabilities
            for vuln in spec.vulnerabilities:
                if vuln.ignored:
                    continue

                vulns_count += 1

                # Determine vulnerability severity
                severity = severity = VulnerabilitySeverityLabels.UNKNOWN
                if (
                    hasattr(vuln, "CVE")
                    and vuln.CVE
                    and hasattr(vuln.CVE, "cvssv3")
                    and vuln.CVE.cvssv3
                ):
                    severity_str = vuln.CVE.cvssv3.get("base_severity", "none").lower()
                    severity = VulnerabilitySeverityLabels(severity_str)

                # Count based on severity
                if severity is VulnerabilitySeverityLabels.CRITICAL:
                    current_critical += 1
                elif severity is VulnerabilitySeverityLabels.HIGH:
                    current_high += 1
                elif severity is VulnerabilitySeverityLabels.MEDIUM:
                    current_medium += 1
                elif severity is VulnerabilitySeverityLabels.LOW:
                    current_low += 1
                else:
                    current_others += 1

            # Check for available fixes - safely access remediation attributes
            if spec.remediation:
                # Access remediation properties safely without relying on specific attribute names
                remediation: RemediationModel = spec.remediation
                has_recommended_version = True if remediation.recommended else False

                if has_recommended_version:
                    current_fixes += 1
                    current_resolved_vulns += len(
                        [v for v in spec.vulnerabilities if not v.ignored]
                    )

        # Update total counts
        critical_vulns_count += current_critical
        high_vulns_count += current_high
        medium_vulns_count += current_medium
        low_vulns_count += current_low
        others_vulns_count += current_others
        fixes_count += current_fixes
        total_resolved_vulns += current_resolved_vulns

        # Save file data for further processing
        file = FileModel(
            location=path,
            file_type=analyzed_file.file_type,
            results=analyzed_file.dependency_results,
        )
        files.append(file)

        # Yield current analysis results
        yield ProgressScanResult(
            type=ScanResultType.PROGRESS,
            percent=progress,
            dependencies=count,
            critical=critical_vulns_count,
            high=high_vulns_count,
            medium=medium_vulns_count,
            low=low_vulns_count,
            others=others_vulns_count,
            vulns_count=vulns_count,
            fixes=fixes_count,
            fixed_vulns=total_resolved_vulns,
            file=str(path),
            file_pkg_count=file_pkg_count,
            file_count=file_count,
            venv_count=venv_count,
        )

    # All files processed, create the report
    project.files = files

    yield StatusScanResult(
        type=ScanResultType.STATUS,
        message="Creating final report",
        action="reporting",
        percent=90,
    )

    # Convert dictionaries to model objects if needed
    if isinstance(metadata, dict):
        metadata_model = MetadataModel(**metadata)
    else:
        metadata_model = metadata

    if isinstance(telemetry, dict):
        telemetry_model = TelemetryModel(**telemetry)
    else:
        telemetry_model = telemetry

    report = ReportModel(
        version=ReportSchemaVersion.v3_0,
        metadata=metadata_model,
        telemetry=telemetry_model,
        files=[],
        projects=[project],
    )

    # Emit uploading status before starting upload
    yield UploadingScanResult(
        type=ScanResultType.UPLOADING, message="Preparing to upload scan results"
    )

    # TODO: Decouple platform upload logic
    try:
        # Convert report to JSON format
        yield UploadingScanResult(
            type=ScanResultType.UPLOADING,
            message="Converting report to JSON format",
            percent=25,
        )
        json_format = report.as_v30().json()

        # Start upload
        yield UploadingScanResult(
            type=ScanResultType.UPLOADING,
            message="Uploading results to Safety platform",
            percent=50,
        )
        result = ctx.obj.auth.client.upload_report(json_format)

        # Upload complete
        yield UploadingScanResult(
            type=ScanResultType.UPLOADING,
            message="Upload completed successfully",
            percent=100,
        )

        scan_id = result.get("uuid")

        codebase_url = f"{SAFETY_PLATFORM_URL}{result['url']}"

    except Exception as e:
        # Emit error status
        yield UploadingScanResult(
            type=ScanResultType.UPLOADING, message=f"Error uploading results: {str(e)}"
        )
        raise e

    # Final yield with completed flag
    yield CompleteScanResult(
        type=ScanResultType.COMPLETE,
        dependencies=count,
        critical=critical_vulns_count,
        high=high_vulns_count,
        medium=medium_vulns_count,
        low=low_vulns_count,
        others=others_vulns_count,
        vulns_count=vulns_count,
        fixes=fixes_count,
        fixed_vulns=total_resolved_vulns,
        codebase_url=codebase_url,
        scan_id=scan_id,
    )

    # Return the complete report and files
    return report, files


def start_scan(
    ctx: "typer.Context",
    auth_type: AuthenticationType,
    is_authenticated: bool,
    target: Path,
    client: Any,
    project: ProjectModel,
    branch: Optional[str] = None,
    stage: Stage = Stage.development,
    platform_enabled: bool = False,
    telemetry_enabled: bool = True,
    use_server_matching: bool = False,
) -> Generator["ScanResult", None, Tuple["ReportModel", List["FileModel"]]]:
    """
    Initialize and start a scan, returning an iterator that yields scan results.
    This function handles setting up all required parameters for the scan.

    Args:
        ctx: The Typer context object containing configuration and project information
        target: The target directory to scan
        use_server_matching: Whether to use server-side vulnerability matching

    Returns:
        An iterator that yields scan results
    """
    if not branch:
        if git_data := GIT(root=target).build_git_data():
            branch = git_data.branch

    command_name = "scan"
    telemetry = build_telemetry_data(
        telemetry=telemetry_enabled, command=command_name, subcommand=None
    )

    scan_type = ScanType(command_name)
    targets = [target]

    if not scan_type:
        raise SafetyException("Missing scan_type.")

    metadata = MetadataModel(
        scan_type=scan_type,
        stage=stage,
        scan_locations=targets,
        authenticated=is_authenticated,
        authentication_type=auth_type,
        telemetry=telemetry,
        schema_version=ReportSchemaVersion.v3_0,
    )

    policy_file_path = target / Path(".safety-policy.yml")

    # Load Policy file and pull it from CLOUD
    local_policy = load_policy_file(policy_file_path)

    cloud_policy = None
    if platform_enabled:
        cloud_policy = download_policy(
            client, project_id=project.id, stage=stage, branch=branch
        )

    project.policy = resolve_policy(local_policy, cloud_policy)
    config = (
        project.policy.config
        if project.policy and project.policy.config
        else ConfigModel()
    )

    return init_scan(
        ctx=ctx,
        target=target,
        config=config,
        metadata=metadata,
        telemetry=telemetry,
        project=project,
        use_server_matching=use_server_matching,
    )



import sys
import logging


logger = logging.getLogger(__name__)


def apply_asyncio_patch():
    """
    Apply a patch to asyncio's exception handling for subprocesses.

    There are some issues with asyncio's exception handling for subprocesses,
    which causes a RuntimeError to be raised when the event loop was already closed.

    This patch catches the RuntimeError and ignores it, which allows the event loop
    to be closed properly.

    Similar issues:
    - https://bugs.python.org/issue39232
    - https://github.com/python/cpython/issues/92841
    """

    import asyncio.base_subprocess

    original_subprocess_del = asyncio.base_subprocess.BaseSubprocessTransport.__del__

    def patched_subprocess_del(self):
        try:
            original_subprocess_del(self)
        except (RuntimeError, ValueError, OSError) as e:
            if isinstance(e, RuntimeError) and str(e) != "Event loop is closed":
                raise
            if isinstance(e, ValueError) and str(e) != "I/O operation on closed pipe":
                raise
            if isinstance(e, OSError) and "[WinError 6]" not in str(e):
                raise
            logger.debug(f"Patched {original_subprocess_del}")

    asyncio.base_subprocess.BaseSubprocessTransport.__del__ = patched_subprocess_del

    if sys.platform == "win32":
        import asyncio.proactor_events as proactor_events

        original_pipe_del = proactor_events._ProactorBasePipeTransport.__del__

        def patched_pipe_del(self):
            try:
                original_pipe_del(self)
            except (RuntimeError, ValueError) as e:
                if isinstance(e, RuntimeError) and str(e) != "Event loop is closed":
                    raise
                if (
                    isinstance(e, ValueError)
                    and str(e) != "I/O operation on closed pipe"
                ):
                    raise
                logger.debug(f"Patched {original_pipe_del}")

        original_repr = proactor_events._ProactorBasePipeTransport.__repr__

        def patched_repr(self):
            try:
                return original_repr(self)
            except ValueError as e:
                if str(e) != "I/O operation on closed pipe":
                    raise
                logger.debug(f"Patched {original_repr}")
                return f"<{self.__class__} [closed]>"

        proactor_events._ProactorBasePipeTransport.__del__ = patched_pipe_del
        proactor_events._ProactorBasePipeTransport.__repr__ = patched_repr

        import subprocess

        original_popen_del = subprocess.Popen.__del__

        def patched_popen_del(self):
            try:
                original_popen_del(self)
            except OSError as e:
                if "[WinError 6]" not in str(e):
                    raise
                logger.debug(f"Patched {original_popen_del}")

        subprocess.Popen.__del__ = patched_popen_del


apply_asyncio_patch()



import random
import time
from rich.text import Text
from rich.style import Style

# -----------------------------
# Celebration Effects
# -----------------------------

def show_confetti(console):
    # Characters to use as confetti
    chars = ['*', 'o', '+', '~']
    width = console.size.width
    height = console.size.height

    frames = 10
    for _ in range(frames):
        console.clear()
        for __ in range(random.randint(50, 100)):  # number of confetti pieces
            x = random.randint(0, max(0, width - 1))
            y = random.randint(0, max(0, height - 2))
            char = random.choice(chars)
            color = random.choice(["red", "green", "yellow", "blue", "magenta", "cyan"])
            console.print(Text(char, style=Style(color=color)), end='', style=color, justify="left", overflow="ignore", no_wrap=True, soft_wrap=False)
            console.file.write(f"\x1b[{y};{x}H")  # Move cursor to position
        console.file.flush()
        time.sleep(0.3)
    console.clear()
    console.print("The confetti has settled! Congrats on a clean scan!", style="bold green")

def show_trophy(console):
    """Displays a celebratory trophy with sparkles."""
    trophy = r"""
       ___________
      '._==_==_=_.'
      .-\:      /-.
     | (|:.     |) |
      '-|:.     |-'
        \::.    /
         '::. .'
           ) (
         _.' '._
        `"""""""`
    """
    for _ in range(5):  # Trophy animation
        console.clear()
        sparkles = random.choice(["✨", "🌟", "💫", "🎉"])
        console.print(trophy, style="bold yellow")
        console.print(f"{sparkles} Scan Complete! No vulnerabilities found! {sparkles}", style="bold green", justify="center")
        time.sleep(0.5)
    console.print("Your code is SAFE and SOUND! 🏆", style="bold yellow")


def show_balloons(console):
    """Displays celebratory balloons popping."""
    balloons = ["🎈", "🎉", "✨", "💥"]
    width = console.size.width

    for _ in range(10):  # Number of balloons
        console.clear()
        for __ in range(random.randint(5, 10)):  # Balloons per frame
            x = random.randint(0, width - 1)
            balloon = random.choice(balloons)
            console.print(Text(balloon, style=Style(color="yellow")), end='', overflow="ignore")
            console.file.write(f"\x1b[{random.randint(1, 10)};{x}H")
        console.file.flush()
        time.sleep(0.5)
    console.print("🎈 POP! 🎉 No vulnerabilities detected!", style="bold green")


def show_victory_parade(console):
    """Displays a victory parade of emojis."""
    parade = ["🎉", "🎊", "🏆", "🥳", "✨", "💻", "🔒", "✅"]
    width = console.size.width

    for _ in range(20):  # Duration of parade
        console.clear()
        line = " ".join(random.choices(parade, k=width // 2))
        console.print(line, style="bold green", justify="center")
        time.sleep(0.2)

    console.print("The parade is over. Your code is safe! 🏆", style="bold yellow")


def show_confetti_rain(console):
    """Displays a colorful confetti rain effect."""
    colors = ["red", "green", "yellow", "blue", "magenta", "cyan", "white"]
    width = console.size.width

    for _ in range(10):  # Number of confetti frames
        console.clear()
        for __ in range(100):  # Confetti pieces per frame
            x = random.randint(0, width - 1)
            char = random.choice(["*", "+", "~", "✨", "o"])
            color = random.choice(colors)
            console.print(Text(char, style=Style(color=color)), end='', overflow="ignore")
            console.file.write(f"\x1b[{random.randint(1, 10)};{x}H")
        console.file.flush()
        time.sleep(0.3)

    console.print("🎉 Confetti celebration complete! You're vulnerability-free! 🎉", style="bold cyan")


def show_fireworks_display(console):
    """Displays a celebratory fireworks animation."""
    fireworks = ["💥", "✨", "🌟", "🎆", "🎇"]
    width = console.size.width

    for _ in range(15):  # Number of fireworks
        x = random.randint(5, width - 5)
        y = random.randint(2, 8)
        firework = random.choice(fireworks)
        color = random.choice(["red", "yellow", "green", "blue", "magenta"])
        console.print(Text(firework, style=Style(color=color)), end='', overflow="ignore")
        console.file.write(f"\x1b[{y};{x}H")  # Position fireworks
        console.file.flush()
        time.sleep(0.3)

    console.print("🎆 Fireworks display finished! Code is secure! 🎆", style="bold magenta")


def show_star_trail(console):
    """Displays a shooting star trail effect."""
    stars = ["⭐", "🌟", "✨", "💫"]
    width = console.size.width

    for _ in range(10):  # Number of shooting stars
        console.clear()
        start_x = random.randint(0, width // 2)
        trail = "".join(random.choices(stars, k=10))
        console.print(f"{' ' * start_x}{trail}", style="bold yellow", justify="left")
        time.sleep(0.3)

    console.print("✨ Your code shines bright with no vulnerabilities! ✨", style="bold cyan")


def show_celebration_wave(console):
    """Displays a celebratory wave effect with emojis."""
    emojis = ["🎉", "🎊", "✨", "🥳", "🎈"]
    width = console.size.width
    wave = [random.choice(emojis) for _ in range(width)]

    for _ in range(10):  # Number of waves
        console.clear()
        line = "".join(wave)
        console.print(line, style="bold yellow", justify="center")
        wave.insert(0, wave.pop())  # Shift wave
        time.sleep(0.3)

    console.print("🌊 Celebration wave ends! Your scan is clean! 🌟", style="bold green")


# List of all celebratory effects
CELEBRATION_EFFECTS = [
    show_confetti,
    show_trophy,
    show_balloons,
    show_victory_parade,
    show_confetti_rain,
    show_fireworks_display,
    show_star_trail,
    show_celebration_wave,
]



import os

from unittest.mock import Mock, patch
from safety.scan.finder.handlers import PythonFileHandler

@patch('safety.safety.fetch_database')
def test_download_required_assets(mock_fetch_database):
    handler = PythonFileHandler()
    session = Mock()

    os.environ["SAFETY_DB_DIR"] = "/path/to/db"
    handler.download_required_assets(session)

    _, kwargs = mock_fetch_database.call_args

    assert kwargs['db'] == "/path/to/db"

@patch('safety.safety.fetch_database')
def test_download_required_assets_no_db_dir(mock_fetch_database):
    handler = PythonFileHandler()
    session = Mock()

    if "SAFETY_DB_DIR" in os.environ:
        del os.environ["SAFETY_DB_DIR"]
    handler.download_required_assets(session)

    _, kwargs = mock_fetch_database.call_args

    assert kwargs['db'] == False



from dataclasses import dataclass, field
from enum import Enum, auto
from typing import Any, Dict, List, Optional, Set


class ToolIntentionType(Enum):
    """
    High-level intentions that are common across tools
    """

    ADD_PACKAGE = auto()
    REMOVE_PACKAGE = auto()
    UPDATE_PACKAGE = auto()
    SYNC_PACKAGES = auto()
    LIST_PACKAGES = auto()
    INIT_PROJECT = auto()
    BUILD_PROJECT = auto()
    RUN_SCRIPT = auto()
    UNKNOWN = auto()


@dataclass
class Dependency:
    """
    Common representation of a dependency
    """

    name: str
    arg_index: int
    original_text: str
    version_constraint: Optional[str] = None
    extras: Set[str] = field(default_factory=set)
    is_dev_dependency: bool = False
    corrected_text: Optional[str] = None


@dataclass
class CommandToolIntention:
    """
    Represents a parsed tool command with normalized intention
    """

    tool: str
    command: str
    intention_type: ToolIntentionType
    packages: List[Dependency] = field(default_factory=list)
    options: Dict[str, Any] = field(default_factory=dict)
    raw_args: List[str] = field(default_factory=list)



from importlib.metadata import distributions
import tomli
from packaging.requirements import Requirement
from packaging.utils import canonicalize_name

def get_project_dependencies():
    try:
        with open("pyproject.toml", "rb") as f:
            pyproject = tomli.load(f)
            deps = pyproject.get("project", {}).get("dependencies", [])
            return {canonicalize_name(Requirement(dep).name): dep 
                   for dep in deps if isinstance(dep, str)}
    except Exception as e:
        print(f"Error reading dependencies: {e}")
        return {}

def pytest_configure(config):
    main_deps_specs = get_project_dependencies()
    all_dists = {canonicalize_name(dist.metadata['Name']): 
                 (dist.metadata['Name'], dist.version) 
                 for dist in distributions()}
    
    # Main dependencies table
    print(f"\n[{len(main_deps_specs)}] Main Dependencies:")
    print("-" * 60)
    print("%-20s %-25s %-15s" % ("Package", "Specification", "Installed"))
    print("-" * 60)
    
    for pkg_norm, spec in sorted(main_deps_specs.items()):
        if pkg_norm in all_dists:
            name, version = all_dists[pkg_norm]
            print("%-20s %-25s %-15s" % (name, spec, version))
    
    other_pkgs = [f"{name} {ver}" 
                  for pkg_norm, (name, ver) in sorted(all_dists.items()) 
                  if pkg_norm not in main_deps_specs]
    
    # Other dependencies in wrapped format
    print(f"\n[{len(other_pkgs)}] Other Dependencies:")
    print("-" * 80)    
    
    # Print other dependencies with wrapping
    line = ""
    for pkg in other_pkgs:
        if len(line) + len(pkg) + 2 > 78:
            print(line.rstrip(", "))
            line = pkg + ", "
        else:
            line += pkg + ", "
    if line:
        print(line.rstrip(", "))


import logging
import subprocess
import sys
from collections import defaultdict
from enum import Enum
from functools import wraps

import time
from typing import (
    TYPE_CHECKING,
    Any,
    DefaultDict,
    Dict,
    List,
    Optional,
    Tuple,
    Union,
    cast,
)

import click
import typer
from rich.console import Console
from rich.table import Table
from rich.text import Text
from typer.core import MarkupMode, TyperCommand, TyperGroup
from click.utils import make_str


from safety.constants import (
    BETA_PANEL_DESCRIPTION_HELP,
    MSG_NO_AUTHD_CICD_PROD_STG,
    MSG_NO_AUTHD_CICD_PROD_STG_ORG,
    MSG_NO_AUTHD_DEV_STG,
    MSG_NO_AUTHD_DEV_STG_ORG_PROMPT,
    MSG_NO_AUTHD_DEV_STG_PROMPT,
    MSG_NO_AUTHD_NOTE_CICD_PROD_STG_TPL,
    MSG_NO_VERIFIED_EMAIL_TPL,
    CONTEXT_COMMAND_TYPE,
    FeatureType,
)
from safety.scan.constants import CONSOLE_HELP_THEME
from safety.models import SafetyCLI

if TYPE_CHECKING:
    from click.core import Command, Context
    from safety.auth.models import Auth


LOG = logging.getLogger(__name__)


class CommandType(Enum):
    MAIN = "main"
    UTILITY = "utility"
    BETA = "beta"


def custom_print_options_panel(
    name: str, params: List[Any], ctx: Any, console: Console
) -> None:
    """
    Print a panel with options.

    Args:
        name (str): The title of the panel.
        params (List[Any]): The list of options/arguments to print.
        ctx (Any): The context object.
        markup_mode (str): The markup mode.
        console (Console): The console to print to.
    """
    table = Table(title=name, show_lines=True)
    for param in params:
        opts = getattr(param, "opts", "")
        help_text = getattr(param, "help", "")
        table.add_row(str(opts), help_text)
    console.print(table)


def custom_print_commands_panel(
    name: str, commands: List[Any], console: Console
) -> None:
    """
    Print a panel with commands.

    Args:
        name (str): The title of the panel.
        commands (List[Any]): The list of commands to print.
        console (Console): The console to print to.
    """
    table = Table(title=name, show_lines=True)
    for command in commands:
        table.add_row(command.name, command.help or "")
    console.print(table)


def custom_make_rich_text(text: str) -> Text:
    """
    Create rich text.

    Args:
        text (str): The text to format.

    Returns:
        Text: The formatted rich text.
    """
    return Text(text)


def custom_get_help_text(obj: Any) -> Text:
    """
    Get the help text for an object.

    Args:
        obj (Any): The object to get help text for.

    Returns:
        Text: The formatted help text.
    """
    return Text(obj.help)


def custom_make_command_help(help_text: str) -> Text:
    """
    Create rich text for command help.

    Args:
        help_text (str): The help text to format.
        markup_mode (str): The markup mode.

    Returns:
        Text: The formatted rich text.
    """
    return Text(help_text)


def get_command_for(name: str, typer_instance: typer.Typer) -> click.Command:
    """
    Retrieve a command by name from a Typer instance.

    Args:
        name (str): The name of the command.
        typer_instance (typer.Typer): The Typer instance.

    Returns:
        click.Command: The found command.
    """
    single_command = next(
        (
            command
            for command in typer_instance.registered_commands
            if command.name == name
        ),
        None,
    )

    if not single_command:
        raise ValueError("Unable to find the command name.")

    single_command.context_settings = typer_instance.info.context_settings
    click_command = typer.main.get_command_from_info(
        single_command,
        pretty_exceptions_short=typer_instance.pretty_exceptions_short,
        rich_markup_mode=typer_instance.rich_markup_mode,
    )
    if typer_instance._add_completion:
        click_install_param, click_show_param = (
            typer.main.get_install_completion_arguments()
        )
        click_command.params.append(click_install_param)
        click_command.params.append(click_show_param)
    return click_command


def pass_safety_cli_obj(func):
    """
    Decorator to ensure the SafetyCLI object exists for a command.
    """

    @wraps(func)
    def inner(ctx, *args, **kwargs):
        if not ctx.obj:
            ctx.obj = SafetyCLI()

        return func(ctx, *args, **kwargs)

    return inner


def pretty_format_help(
    obj: Union[click.Command, click.Group], ctx: click.Context, markup_mode: MarkupMode
) -> None:
    """
    Format and print help text in a pretty format.

    Args:
        obj (Union[click.Command, click.Group]): The Click command or group.
        ctx (click.Context): The Click context.
        markup_mode (MarkupMode): The markup mode.
    """
    from rich.align import Align
    from rich.console import Console
    from rich.padding import Padding
    from rich.theme import Theme
    from typer.rich_utils import (
        ARGUMENTS_PANEL_TITLE,
        COMMANDS_PANEL_TITLE,
        OPTIONS_PANEL_TITLE,
        STYLE_USAGE_COMMAND,
        highlighter,
    )

    console = Console()

    with console.use_theme(Theme(styles=CONSOLE_HELP_THEME)) as theme_context:
        console = theme_context.console
        # Print command / group help if we have some
        if obj.help:
            console.print()

            # Print with some padding
            console.print(
                Padding(Align(custom_get_help_text(obj=obj), pad=False), (0, 1, 0, 1))
            )

        # Print usage
        console.print(
            Padding(highlighter(obj.get_usage(ctx)), 1), style=STYLE_USAGE_COMMAND
        )

        if isinstance(obj, click.MultiCommand):
            panel_to_commands: DefaultDict[str, List[click.Command]] = defaultdict(list)
            for command_name in obj.list_commands(ctx):
                command = obj.get_command(ctx, command_name)
                if command and not command.hidden:
                    panel_name = (
                        getattr(command, "rich_help_panel", None)
                        or COMMANDS_PANEL_TITLE
                    )
                    panel_to_commands[panel_name].append(command)

            # Print each command group panel
            default_commands = panel_to_commands.get(COMMANDS_PANEL_TITLE, [])
            custom_print_commands_panel(
                name=COMMANDS_PANEL_TITLE,
                commands=default_commands,
                console=console,
            )
            for panel_name, commands in panel_to_commands.items():
                if panel_name == COMMANDS_PANEL_TITLE:
                    # Already printed above
                    continue
                custom_print_commands_panel(
                    name=panel_name,
                    commands=commands,
                    console=console,
                )

        panel_to_arguments: DefaultDict[str, List[click.Argument]] = defaultdict(list)
        panel_to_options: DefaultDict[str, List[click.Option]] = defaultdict(list)
        for param in obj.get_params(ctx):
            # Skip if option is hidden
            if getattr(param, "hidden", False):
                continue
            if isinstance(param, click.Argument):
                panel_name = (
                    getattr(param, "rich_help_panel", None) or ARGUMENTS_PANEL_TITLE
                )
                panel_to_arguments[panel_name].append(param)
            elif isinstance(param, click.Option):
                panel_name = (
                    getattr(param, "rich_help_panel", None) or OPTIONS_PANEL_TITLE
                )
                panel_to_options[panel_name].append(param)

        default_options = panel_to_options.get(OPTIONS_PANEL_TITLE, [])
        custom_print_options_panel(
            name=OPTIONS_PANEL_TITLE,
            params=default_options,
            ctx=ctx,
            console=console,
        )
        for panel_name, options in panel_to_options.items():
            if panel_name == OPTIONS_PANEL_TITLE:
                # Already printed above
                continue
            custom_print_options_panel(
                name=panel_name,
                params=options,
                ctx=ctx,
                console=console,
            )

        default_arguments = panel_to_arguments.get(ARGUMENTS_PANEL_TITLE, [])
        custom_print_options_panel(
            name=ARGUMENTS_PANEL_TITLE,
            params=default_arguments,
            ctx=ctx,
            console=console,
        )
        for panel_name, arguments in panel_to_arguments.items():
            if panel_name == ARGUMENTS_PANEL_TITLE:
                # Already printed above
                continue
            custom_print_options_panel(
                name=panel_name,
                params=arguments,
                ctx=ctx,
                console=console,
            )

        if ctx.parent:
            params = []
            for param in ctx.parent.command.params:
                if isinstance(param, click.Option):
                    params.append(param)

            custom_print_options_panel(
                name="Global-Options",
                params=params,
                ctx=ctx.parent,
                console=console,
            )

        # Epilogue if we have it
        if obj.epilog:
            # Remove single linebreaks, replace double with single
            lines = obj.epilog.split("\n\n")
            epilogue = "\n".join([x.replace("\n", " ").strip() for x in lines])
            epilogue_text = custom_make_rich_text(text=epilogue)
            console.print(Padding(Align(epilogue_text, pad=False), 1))


def print_main_command_panels(
    *,
    name: str,
    commands_type: CommandType,
    commands: List[click.Command],
    markup_mode: MarkupMode,
    console,
) -> None:
    """
    Print the main command panels.

    Args:
        name (str): The name of the panel.
        commands (List[click.Command]): List of commands to display.
        markup_mode (MarkupMode): The markup mode.
        console: The Rich console.
    """
    from rich import box
    from rich.panel import Panel
    from rich.table import Table
    from rich.text import Text
    from typer.rich_utils import (
        ALIGN_COMMANDS_PANEL,
        STYLE_COMMANDS_PANEL_BORDER,
        STYLE_COMMANDS_TABLE_BORDER_STYLE,
        STYLE_COMMANDS_TABLE_BOX,
        STYLE_COMMANDS_TABLE_LEADING,
        STYLE_COMMANDS_TABLE_PAD_EDGE,
        STYLE_COMMANDS_TABLE_PADDING,
        STYLE_COMMANDS_TABLE_ROW_STYLES,
        STYLE_COMMANDS_TABLE_SHOW_LINES,
    )

    t_styles: Dict[str, Any] = {
        "show_lines": STYLE_COMMANDS_TABLE_SHOW_LINES,
        "leading": STYLE_COMMANDS_TABLE_LEADING,
        "box": STYLE_COMMANDS_TABLE_BOX,
        "border_style": STYLE_COMMANDS_TABLE_BORDER_STYLE,
        "row_styles": STYLE_COMMANDS_TABLE_ROW_STYLES,
        "pad_edge": STYLE_COMMANDS_TABLE_PAD_EDGE,
        "padding": STYLE_COMMANDS_TABLE_PADDING,
    }
    box_style = getattr(box, t_styles.pop("box"), None)

    commands_table = Table(
        highlight=False,
        show_header=False,
        expand=True,
        box=box_style,
        **t_styles,
    )

    console_width = 80
    column_width = 25

    if console.size and console.size[0] > 80:
        console_width = console.size[0]

    from rich.console import Group

    description = None

    if commands_type is CommandType.BETA:
        description = Group(Text(""), Text(BETA_PANEL_DESCRIPTION_HELP), Text(""))

    commands_table.add_column(
        style="bold cyan", no_wrap=True, width=column_width, max_width=column_width
    )
    commands_table.add_column(width=console_width - column_width)

    rows = []

    for command in commands:
        helptext = command.short_help or command.help or ""
        command_name = command.name or ""
        command_name_text = (
            Text(command_name, style="")
            if commands_type is CommandType.BETA
            else Text(command_name)
        )
        rows.append(
            [
                command_name_text,
                custom_make_command_help(
                    help_text=helptext,
                ),
            ]
        )
        rows.append([])
    for row in rows:
        commands_table.add_row(*row)
    if commands_table.row_count:
        renderables = (
            [description, commands_table]
            if description is not None
            else [Text(""), commands_table]
        )

        console.print(
            Panel(
                Group(*renderables),
                border_style=STYLE_COMMANDS_PANEL_BORDER,
                title=name,
                title_align=ALIGN_COMMANDS_PANEL,
            )
        )


# The help output for the main safety root command: `safety --help`
def format_main_help(
    obj: Union[click.Command, click.Group], ctx: click.Context, markup_mode: MarkupMode
) -> None:
    """
    Format the main help output for the safety root command.

    Args:
        obj (Union[click.Command, click.Group]): The Click command or group.
        ctx (click.Context): The Click context.
        markup_mode (MarkupMode): The markup mode.
    """
    from rich.align import Align
    from rich.console import Console
    from rich.padding import Padding
    from rich.theme import Theme
    from typer.rich_utils import (
        ARGUMENTS_PANEL_TITLE,
        COMMANDS_PANEL_TITLE,
        OPTIONS_PANEL_TITLE,
        STYLE_USAGE_COMMAND,
        highlighter,
    )

    typer_console = Console()

    with typer_console.use_theme(Theme(styles=CONSOLE_HELP_THEME)) as theme_context:
        console = theme_context.console

        # Print command / group help if we have some
        if obj.help:
            console.print()
            # Print with some padding
            console.print(
                Padding(
                    Align(
                        custom_get_help_text(obj=obj),
                        pad=False,
                    ),
                    (0, 1, 0, 1),
                )
            )

        # Print usage
        console.print(
            Padding(highlighter(obj.get_usage(ctx)), 1), style=STYLE_USAGE_COMMAND
        )

        if isinstance(obj, click.MultiCommand):
            UTILITY_COMMANDS_PANEL_TITLE = "Utility commands"
            BETA_COMMANDS_PANEL_TITLE = "Beta Commands :rocket:"

            COMMANDS_PANEL_TITLE_CONSTANTS = {
                CommandType.MAIN: COMMANDS_PANEL_TITLE,
                CommandType.UTILITY: UTILITY_COMMANDS_PANEL_TITLE,
                CommandType.BETA: BETA_COMMANDS_PANEL_TITLE,
            }

            panel_to_commands: Dict[CommandType, List[click.Command]] = {}

            # Keep order of panels
            for command_type in COMMANDS_PANEL_TITLE_CONSTANTS.keys():
                panel_to_commands[command_type] = []

            for command_name in obj.list_commands(ctx):
                command = obj.get_command(ctx, command_name)
                if command and not command.hidden:
                    command_type = command.context_settings.get(
                        CONTEXT_COMMAND_TYPE, CommandType.MAIN
                    )
                    panel_to_commands[command_type].append(command)

            for command_type, commands in panel_to_commands.items():
                print_main_command_panels(
                    name=COMMANDS_PANEL_TITLE_CONSTANTS[command_type],
                    commands_type=command_type,
                    commands=commands,
                    markup_mode=markup_mode,
                    console=console,
                )

        panel_to_arguments: DefaultDict[str, List[click.Argument]] = defaultdict(list)
        panel_to_options: DefaultDict[str, List[click.Option]] = defaultdict(list)
        for param in obj.get_params(ctx):
            # Skip if option is hidden
            if getattr(param, "hidden", False):
                continue
            if isinstance(param, click.Argument):
                panel_name = (
                    getattr(param, "rich_help_panel", None) or ARGUMENTS_PANEL_TITLE
                )
                panel_to_arguments[panel_name].append(param)
            elif isinstance(param, click.Option):
                panel_name = (
                    getattr(param, "rich_help_panel", None) or OPTIONS_PANEL_TITLE
                )
                panel_to_options[panel_name].append(param)
        default_arguments = panel_to_arguments.get(ARGUMENTS_PANEL_TITLE, [])
        custom_print_options_panel(
            name=ARGUMENTS_PANEL_TITLE,
            params=default_arguments,
            ctx=ctx,
            console=console,
        )
        for panel_name, arguments in panel_to_arguments.items():
            if panel_name == ARGUMENTS_PANEL_TITLE:
                # Already printed above
                continue
            custom_print_options_panel(
                name=panel_name,
                params=arguments,
                ctx=ctx,
                console=console,
            )
        default_options = panel_to_options.get(OPTIONS_PANEL_TITLE, [])
        custom_print_options_panel(
            name=OPTIONS_PANEL_TITLE,
            params=default_options,
            ctx=ctx,
            console=console,
        )
        for panel_name, options in panel_to_options.items():
            if panel_name == OPTIONS_PANEL_TITLE:
                # Already printed above
                continue
            custom_print_options_panel(
                name=panel_name,
                params=options,
                ctx=ctx,
                console=console,
            )

        # Epilogue if we have it
        if obj.epilog:
            # Remove single linebreaks, replace double with single
            lines = obj.epilog.split("\n\n")
            epilogue = "\n".join([x.replace("\n", " ").strip() for x in lines])
            epilogue_text = custom_make_rich_text(text=epilogue)
            console.print(Padding(Align(epilogue_text, pad=False), 1))


def process_auth_status_not_ready(console, auth: "Auth", ctx: typer.Context) -> None:
    """
    Handle the process when the authentication status is not ready.

    Args:
        console: The Rich console.
        auth (Auth): The Auth object.
        ctx (typer.Context): The Typer context.
    """
    from rich.prompt import Confirm, Prompt
    from safety_schemas.models import Stage
    from safety.auth.constants import CLI_AUTH, MSG_NON_AUTHENTICATED

    if not auth.client or not auth.client.is_using_auth_credentials():
        if auth.stage is Stage.development:
            console.print()
            if auth.org:
                confirmed = Confirm.ask(
                    MSG_NO_AUTHD_DEV_STG_ORG_PROMPT,
                    choices=["Y", "N", "y", "n"],
                    show_choices=False,
                    show_default=False,
                    default=True,
                    console=console,
                )

                if not confirmed:
                    sys.exit(0)

                from safety.auth.cli import auth_app

                login_command = get_command_for(name="login", typer_instance=auth_app)
                ctx.invoke(login_command)
            else:
                console.print(MSG_NO_AUTHD_DEV_STG)
                console.print()
                choices = ["L", "R", "l", "r"]
                next_command = Prompt.ask(
                    MSG_NO_AUTHD_DEV_STG_PROMPT,
                    default=None,
                    choices=choices,
                    show_choices=False,
                    console=console,
                )

                from safety.auth.cli import auth_app

                login_command = get_command_for(name="login", typer_instance=auth_app)
                register_command = get_command_for(
                    name="register", typer_instance=auth_app
                )
                if next_command is None or next_command.lower() not in choices:
                    sys.exit(0)

                console.print()
                if next_command.lower() == "r":
                    ctx.invoke(register_command)
                else:
                    ctx.invoke(login_command)

                if not ctx.obj.auth.email_verified:
                    sys.exit(1)
        else:
            if not auth.org:
                console.print(MSG_NO_AUTHD_CICD_PROD_STG_ORG.format(LOGIN_URL=CLI_AUTH))

            else:
                console.print(MSG_NO_AUTHD_CICD_PROD_STG)
                console.print(
                    MSG_NO_AUTHD_NOTE_CICD_PROD_STG_TPL.format(
                        LOGIN_URL=CLI_AUTH, SIGNUP_URL=f"{CLI_AUTH}/?sign_up=True"
                    )
                )
            sys.exit(1)

    elif not auth.email_verified:
        console.print()
        console.print(
            MSG_NO_VERIFIED_EMAIL_TPL.format(
                email=auth.email if auth.email else "Missing email"
            )
        )
        sys.exit(1)
    else:
        console.print(MSG_NON_AUTHENTICATED)
        sys.exit(1)


class CustomContext(click.Context):
    def __init__(
        self,
        command: "Command",
        parent: Optional["Context"] = None,
        command_type: CommandType = CommandType.MAIN,
        feature_type: Optional[FeatureType] = None,
        **kwargs,
    ) -> None:
        self.command_type = command_type
        self.feature_type = feature_type
        self.started_at = time.monotonic()
        self.command_alias_used: Optional[str] = None
        super().__init__(command, parent=parent, **kwargs)


class SafetyCLISubGroup(TyperGroup):
    """
    Custom TyperGroup with additional functionality for Safety CLI.
    """

    context_class = CustomContext

    def format_help(self, ctx: click.Context, formatter: click.HelpFormatter) -> None:
        """
        Format help message with rich formatting.

        Args:
            ctx (click.Context): Click context.
            formatter (click.HelpFormatter): Click help formatter.
        """
        pretty_format_help(self, ctx, markup_mode=self.rich_markup_mode)

    def format_usage(self, ctx: click.Context, formatter: click.HelpFormatter) -> None:
        """
        Format usage message.

        Args:
            ctx (click.Context): Click context.
            formatter (click.HelpFormatter): Click help formatter.
        """
        command_path = ctx.command_path
        pieces = self.collect_usage_pieces(ctx)
        main_group = ctx.parent
        if main_group:
            command_path = (
                f"{main_group.command_path} [GLOBAL-OPTIONS] {ctx.command.name}"
            )

        formatter.write_usage(command_path, " ".join(pieces))

    def command(
        self,
        *args: Any,
        **kwargs: Any,
    ) -> click.Command:  # type: ignore[override]
        """
        Create a new command.

        Args:
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.

        Returns:
            click.Command: The created command.
        """
        super().command(*args, **kwargs)


class SafetyCLICommand(TyperCommand):
    """
    Custom TyperCommand with additional functionality for Safety CLI.
    """

    context_class = CustomContext

    def format_help(self, ctx: click.Context, formatter: click.HelpFormatter) -> None:
        """
        Format help message with rich formatting.

        Args:
            ctx (click.Context): Click context.
            formatter (click.HelpFormatter): Click help formatter.
        """
        pretty_format_help(self, ctx, markup_mode=self.rich_markup_mode)

    def format_usage(self, ctx: click.Context, formatter: click.HelpFormatter) -> None:
        """
        Format usage message.

        Args:
            ctx (click.Context): Click context.
            formatter (click.HelpFormatter): Click help formatter.
        """
        command_path = ctx.command_path
        pieces = self.collect_usage_pieces(ctx)
        main_group = ctx.parent
        if main_group:
            command_path = (
                f"{main_group.command_path} [GLOBAL-OPTIONS] {ctx.command.name}"
            )

        formatter.write_usage(command_path, " ".join(pieces))


class SafetyCLILegacyGroup(click.Group):
    """
    Custom Click Group to handle legacy command-line arguments.
    """

    context_class = CustomContext

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.all_commands = {}

    def __call__(self, *args, **kwargs):
        return super().__call__(*args, **kwargs)

    def add_command(self, cmd, name=None) -> None:
        super().add_command(cmd, name)

        name = name or cmd.name
        self.all_commands[name] = cmd

    def parse_args(self, ctx: click.Context, args: List[str]) -> List[str]:
        ctx = cast(CustomContext, ctx)

        if len(args) >= 1:
            if "pip" in args[0] and ctx:
                ctx.command_alias_used = args[0]
                args[0] = "pip"

        parsed_args = super().parse_args(ctx, args)

        args = ctx.args

        # Workaround for legacy check options, that now are global options
        subcommand_args = set(args)
        PROXY_HOST_OPTIONS = set(["--proxy-host", "-ph"])
        if (
            "check" in ctx.protected_args
            or "license" in ctx.protected_args
            and (
                bool(
                    PROXY_HOST_OPTIONS.intersection(subcommand_args)
                    or "--key" in subcommand_args
                )
            )
        ):
            proxy_options, key = self.parse_legacy_args(args)
            if proxy_options:
                ctx.params.update(proxy_options)

            if key:
                ctx.params.update({"key": key})

        return parsed_args

    def parse_legacy_args(
        self, args: List[str]
    ) -> Tuple[Optional[Dict[str, str]], Optional[str]]:
        """
        Parse legacy command-line arguments for proxy settings and keys.

        Args:
            args (List[str]): List of command-line arguments.

        Returns:
            Tuple[Optional[Dict[str, str]], Optional[str]]: Parsed proxy options and key.
        """
        options = {"proxy_protocol": "https", "proxy_port": 80, "proxy_host": None}
        key = None

        for i, arg in enumerate(args):
            if arg in ["--proxy-protocol", "-pr"] and i + 1 < len(args):
                options["proxy_protocol"] = args[i + 1]
            elif arg in ["--proxy-port", "-pp"] and i + 1 < len(args):
                options["proxy_port"] = int(args[i + 1])
            elif arg in ["--proxy-host", "-ph"] and i + 1 < len(args):
                options["proxy_host"] = args[i + 1]
            elif arg in ["--key"] and i + 1 < len(args):
                key = args[i + 1]

        proxy = options if options["proxy_host"] else None
        return proxy, key

    def get_filtered_commands(self, ctx: click.Context) -> Dict[str, click.Command]:
        from safety.auth.utils import initialize

        initialize(ctx, refresh=False)

        # Filter commands here:
        from .constants import CONTEXT_FEATURE_TYPE

        disabled_features = [
            feature_type
            for feature_type in FeatureType
            if not getattr(ctx.obj, feature_type.attr_name, False)
        ]

        return {
            k: v
            for k, v in self.commands.items()
            if v.context_settings.get(CONTEXT_FEATURE_TYPE, None)
            not in disabled_features
            or k in ["firewall"]
        }

    def invoke(self, ctx: click.Context) -> None:
        """
        Invoke the command, handling legacy arguments.

        Args:
            ctx (click.Context): Click context.
        """
        session_kwargs = {
            "ctx": ctx,
            "proxy_protocol": ctx.params.pop("proxy_protocol", None),
            "proxy_host": ctx.params.pop("proxy_host", None),
            "proxy_port": ctx.params.pop("proxy_port", None),
            "key": ctx.params.pop("key", None),
            "stage": ctx.params.pop("stage", None),
        }
        invoked_command = make_str(next(iter(ctx.protected_args), ""))

        from safety.auth.cli_utils import inject_session

        inject_session(**session_kwargs, invoked_command=invoked_command)

        # call initialize if the --key is used.
        if session_kwargs["key"]:
            from safety.auth.utils import initialize

            initialize(ctx, refresh=True)

        self.commands = self.get_filtered_commands(ctx)

        # Now, invoke the original behavior
        super(SafetyCLILegacyGroup, self).invoke(ctx)

    def list_commands(self, ctx: click.Context) -> List[str]:
        """Override click.Group.list_commands with custom filtering"""
        self.commands = self.get_filtered_commands(ctx)

        return super().list_commands(ctx)

    def format_help(self, ctx: click.Context, formatter: click.HelpFormatter) -> None:
        """
        Format help message with rich formatting.

        Args:
            ctx (click.Context): Click context.
            formatter (click.HelpFormatter): Click help formatter.
        """
        # The main `safety --help`
        if self.name == "cli":
            format_main_help(self, ctx, markup_mode="rich")
        # All other help outputs
        else:
            pretty_format_help(self, ctx, markup_mode="rich")


class SafetyCLILegacyCommand(click.Command):
    """
    Custom Click Command to handle legacy command-line arguments.
    """

    context_class = CustomContext

    def format_help(self, ctx: click.Context, formatter: click.HelpFormatter) -> None:
        """
        Format help message with rich formatting.

        Args:
            ctx (click.Context): Click context.
            formatter (click.HelpFormatter): Click help formatter.
        """
        pretty_format_help(self, ctx, markup_mode="rich")


def get_git_branch_name() -> Optional[str]:
    """
    Retrieves the current Git branch name.

    Returns:
        str: The current Git branch name, or None if it cannot be determined.
    """
    try:
        branch_name = subprocess.check_output(
            ["git", "rev-parse", "--abbrev-ref", "HEAD"],
            stderr=subprocess.DEVNULL,
            text=True,
        ).strip()
        return branch_name if branch_name else None
    except Exception:
        return None



from typing import Dict, Generic, TypeVar, Generator

from pydantic import BaseModel as PydanticBaseModel
from pydantic import Extra
from pydantic.validators import dict_validator

from common.const import SCHEMA_DICT_ITEMS_COUNT_LIMIT
from common.exceptions import DictMaxLengthError


class BaseModel(PydanticBaseModel):
    """
    Base model that extends Pydantic's BaseModel with additional configurations.
    """
    class Config:
        arbitrary_types_allowed: bool = True
        max_anystr_length: int = 50
        validate_assignment: bool = True
        extra = Extra.forbid


KeyType = TypeVar("KeyType")
ValueType = TypeVar("ValueType")


class ConstrainedDict(Generic[KeyType, ValueType]):
    """
    A constrained dictionary that validates its length based on a specified limit.
    """
    def __init__(self, v: Dict[KeyType, ValueType]) -> None:
        """
        Initialize the ConstrainedDict.

        Args:
            v (Dict[KeyType, ValueType]): The dictionary to constrain.
        """
        super().__init__()

    @classmethod
    def __get_validators__(cls) -> Generator:
        yield cls.dict_length_validator

    @classmethod
    def dict_length_validator(cls, v: Dict[KeyType, ValueType]) -> Dict[KeyType, ValueType]:
        """
        Validate the length of the dictionary.

        Args:
            v (Dict[KeyType, ValueType]): The dictionary to validate.

        Returns:
            Dict[KeyType, ValueType]: The validated dictionary.

        Raises:
            DictMaxLengthError: If the dictionary exceeds the allowed length.
        """
        v = dict_validator(v)
        if len(v) > SCHEMA_DICT_ITEMS_COUNT_LIMIT:
            raise DictMaxLengthError(limit_value=SCHEMA_DICT_ITEMS_COUNT_LIMIT)
        return v



# -*- coding: utf-8 -*-
from dataclasses import asdict
import errno
import itertools
import json
import logging
import os
from pathlib import Path
import sys
import tempfile
import time
from collections import defaultdict
from datetime import datetime
from typing import Dict, Optional, List, Any, Union, Iterator

import click
import requests
from packaging.specifiers import SpecifierSet
from packaging.utils import canonicalize_name
from packaging.version import parse as parse_version, Version
from pydantic.json import pydantic_encoder
from filelock import FileLock

from safety_schemas.models import Ecosystem, FileType



from .constants import (API_MIRRORS, DB_CACHE_FILE, OPEN_MIRRORS, REQUEST_TIMEOUT, DATA_API_BASE_URL, JSON_SCHEMA_VERSION,
                        IGNORE_UNPINNED_REQ_REASON)
from .errors import (DatabaseFetchError, DatabaseFileNotFoundError, InvalidCredentialError,
                     TooManyRequestsError, NetworkConnectionError,
                     RequestTimeoutError, ServerError, MalformedDatabase)
from .models import Vulnerability, CVE, Severity, Fix, is_pinned_requirement, SafetyRequirement
from .output_utils import print_service, get_applied_msg, prompt_service, get_skipped_msg, get_fix_opt_used_msg, \
    is_using_api_key, get_specifier_range_info
from .util import build_remediation_info_url, pluralize, read_requirements, Package, build_telemetry_data, sync_safety_context, \
    SafetyContext, validate_expiration_date, is_a_remote_mirror, get_requirements_content, SafetyPolicyFile, \
    get_terminal_size, is_ignore_unpinned_mode, get_hashes

LOG = logging.getLogger(__name__)


def get_from_cache(db_name: str, cache_valid_seconds: int = 0, skip_time_verification: bool = False) -> Optional[Dict[str, Any]]:
    """
    Retrieves the database from the cache if it is valid.

    Args:
        db_name (str): The name of the database.
        cache_valid_seconds (int): The validity period of the cache in seconds.
        skip_time_verification (bool): Whether to skip time verification.

    Returns:
        Optional[[Dict[str, Any]]: The cached database if available and valid, otherwise False.
    """
    cache_file_lock = f"{DB_CACHE_FILE}.lock"
    os.makedirs(os.path.dirname(cache_file_lock), exist_ok=True)
    lock = FileLock(cache_file_lock, timeout=10)
    with lock:
        if os.path.exists(DB_CACHE_FILE):
            with open(DB_CACHE_FILE) as f:
                try:
                    data = json.loads(f.read())
                    if db_name in data:

                        if "cached_at" in data[db_name]:
                            if data[db_name]["cached_at"] + cache_valid_seconds > time.time() or skip_time_verification:
                                LOG.debug('Getting the database from cache at %s, cache setting: %s',
                                        data[db_name]["cached_at"], cache_valid_seconds)

                                try:
                                    data[db_name]["db"]["meta"]["base_domain"] = "https://data.safetycli.com"
                                except KeyError as e:
                                    pass

                                return data[db_name]["db"]

                            LOG.debug('Cached file is too old, it was cached at %s', data[db_name]["cached_at"])
                        else:
                            LOG.debug('There is not the cached_at key in %s database', data[db_name])

                except json.JSONDecodeError:
                    LOG.debug('JSONDecodeError trying to get the cached database.')
        else:
            LOG.debug("Cache file doesn't exist...")
    return None


def write_to_cache(db_name: str, data: Dict[str, Any]) -> None:
    """
    Writes the database to the cache.

    Args:
        db_name (str): The name of the database.
        data (Dict[str, Any]): The database data to be cached.
    """
    # cache is in: ~/safety/cache.json
    # and has the following form:
    # {
    #   "insecure.json": {
    #       "cached_at": 12345678
    #       "db": {}
    #   },
    #   "insecure_full.json": {
    #       "cached_at": 12345678
    #       "db": {}
    #   },
    # }
    if not os.path.exists(os.path.dirname(DB_CACHE_FILE)):
        try:
            os.makedirs(os.path.dirname(DB_CACHE_FILE))
            with open(DB_CACHE_FILE, "w") as _:
                _.write(json.dumps({}))
                LOG.debug('Cache file created')
        except OSError as exc:  # Guard against race condition
            LOG.debug('Unable to create the cache file because: %s', exc.errno)
            if exc.errno != errno.EEXIST:
                raise

    cache_file_lock = f"{DB_CACHE_FILE}.lock"
    lock = FileLock(cache_file_lock, timeout=10)
    with lock:
        if os.path.exists(DB_CACHE_FILE):
            with open(DB_CACHE_FILE, "r") as f:
                try:
                    cache = json.loads(f.read())
                except json.JSONDecodeError:
                    LOG.debug('JSONDecodeError in the local cache, dumping the full cache file.')
                    cache = {}
        else:
            cache = {}

        with open(DB_CACHE_FILE, "w") as f:
            cache[db_name] = {
                "cached_at": time.time(),
                "db": data
            }
            f.write(json.dumps(cache))
            LOG.debug('Safety updated the cache file for %s database.', db_name)


def fetch_database_url(
    session: requests.Session,
    mirror: str,
    db_name: str,
    cached: int,
    telemetry: bool = True,
    ecosystem: Ecosystem = Ecosystem.PYTHON,
    from_cache: bool = True
) -> Dict[str, Any]:
    """
    Fetches the database from a URL.

    Args:
        session (requests.Session): The requests session.
        mirror (str): The URL of the mirror.
        db_name (str): The name of the database.
        cached (int): The cache validity in seconds.
        telemetry (bool): Whether to include telemetry data.
        ecosystem (Ecosystem): The ecosystem.
        from_cache (bool): Whether to fetch from cache.

    Returns:
        Dict[str, Any]: The fetched database.
    """
    headers = {'schema-version': JSON_SCHEMA_VERSION, 'ecosystem': ecosystem.value}

    if cached and from_cache:
        cached_data = get_from_cache(db_name=db_name, cache_valid_seconds=cached)
        if cached_data:
            LOG.info('Database %s returned from cache.', db_name)
            return cached_data
    url = mirror + db_name


    telemetry_data = {
        'telemetry': json.dumps(build_telemetry_data(telemetry=telemetry),
                                default=pydantic_encoder)}

    try:
        r = session.get(url=url, timeout=REQUEST_TIMEOUT,
                        headers=headers, params=telemetry_data)
    except requests.exceptions.ConnectionError:
        raise NetworkConnectionError()
    except requests.exceptions.Timeout:
        raise RequestTimeoutError()
    except requests.exceptions.RequestException:
        raise DatabaseFetchError()

    if r.status_code == 403:
        raise InvalidCredentialError(credential=session.get_credential(), reason=r.text)

    if r.status_code == 429:
        raise TooManyRequestsError(reason=r.text)

    if r.status_code != 200:
        raise ServerError(reason=r.reason)

    try:
        data = r.json()
    except json.JSONDecodeError as e:
        raise MalformedDatabase(reason=e)

    if cached:
        LOG.info('Writing %s to cache because cached value was %s', db_name, cached)
        write_to_cache(db_name, data)

    return data


def fetch_policy(session: requests.Session) -> Dict[str, Any]:
    """
    Fetches the policy from the server.

    Args:
        session (requests.Session): The requests session.

    Returns:
        Dict[str, Any]: The fetched policy.
    """
    url = f"{DATA_API_BASE_URL}policy/"

    try:
        LOG.debug(f'Getting policy')
        r = session.get(url=url, timeout=REQUEST_TIMEOUT)
        LOG.debug(r.text)
        return r.json()
    except Exception:
        LOG.exception("Error fetching policy")

        return {"safety_policy": "", "audit_and_monitor": False}


def post_results(session: requests.Session, safety_json: str, policy_file: str) -> Dict[str, Any]:
    """
    Posts the scan results to the server.

    Args:
        session (requests.Session): The requests session.
        safety_json (str): The scan results in JSON format.
        policy_file (str): The policy file.

    Returns:
        Dict[str, Any]: The server response.
    """
    url = f"{DATA_API_BASE_URL}result/"

    # safety_json is in text form already. policy_file is a text YAML
    audit_report = {
        "safety_json": json.loads(safety_json),
        "policy_file": policy_file
    }

    try:
        LOG.debug(f'Posting results to: {url}')
        LOG.debug(f'Posting results: {audit_report}')
        r = session.post(url=url, timeout=REQUEST_TIMEOUT, json=audit_report)
        LOG.debug(r.text)

        return r.json()
    except:
        LOG.exception("Error posting results")
        click.secho(
            "Warning: couldn't upload results to safetycli.com.",
            fg="yellow",
            file=sys.stderr
        )

        return {}


def fetch_database_file(path: str, db_name: str, cached: int = 0, ecosystem: Optional[Ecosystem] = None) -> Dict[str, Any]:
    """
    Fetches the database from a local file.

    Args:
        path (str): The path to the local file.
        db_name (str): The name of the database.
        cached (int): The cache validity in seconds.
        ecosystem (Optional[Ecosystem]): The ecosystem.

    Returns:
        Dict[str, Any]: The fetched database.
    """
    full_path = (Path(path) / (ecosystem.value if ecosystem else '') / db_name).expanduser().resolve()

    if not full_path.exists():
        raise DatabaseFileNotFoundError(db=path)

    with open(full_path) as f:
        data = json.loads(f.read())

    if cached:
        LOG.info('Writing %s to cache because cached value was %s', db_name, cached)
        write_to_cache(db_name, data)

    return data


def is_valid_database(db: Dict[str, Any]) -> bool:
    """
    Checks if the database is valid.

    Args:
        db (Dict[str, Any]): The database.

    Returns:
        bool: True if the database is valid, False otherwise.
    """
    try:
        if db['meta']['schema_version'] == JSON_SCHEMA_VERSION:
            return True
    except Exception:
        return False

    return False


def fetch_database(
    session: requests.Session,
    full: bool = False,
    db: Union[Optional[str], bool] = False,
    cached: int = 0,
    telemetry: bool = True,
    ecosystem: Optional[Ecosystem] = None,
    from_cache: bool = True
) -> Dict[str, Any]:
    """
    Fetches the database from a mirror or a local file.

    Args:
        session (requests.Session): The requests session.
        full (bool): Whether to fetch the full database.
        db (Optional[str]): The path to the local database file.
        cached (int): The cache validity in seconds.
        telemetry (bool): Whether to include telemetry data.
        ecosystem (Optional[Ecosystem]): The ecosystem.
        from_cache (bool): Whether to fetch from cache.

    Returns:
        Dict[str, Any]: The fetched database.
    """
    if session.is_using_auth_credentials():
        mirrors = API_MIRRORS
    elif db:
        mirrors = [db]
    else:
        mirrors = OPEN_MIRRORS

    db_name = "insecure_full.json" if full else "insecure.json"
    for mirror in mirrors:
        # mirror can either be a local path or a URL
        if is_a_remote_mirror(mirror):
            if ecosystem is None:
                ecosystem = Ecosystem.PYTHON
            data = fetch_database_url(session, mirror, db_name=db_name, cached=cached,
                                      telemetry=telemetry, ecosystem=ecosystem, from_cache=from_cache)
        else:
            data = fetch_database_file(mirror, db_name=db_name, cached=cached,
                                       ecosystem=ecosystem)
        if data:
            if is_valid_database(data):
                return data
            raise MalformedDatabase(fetched_from=mirror,
                                    reason=f'Not supported schema version. '
                                           f'This Safety version supports only schema version {JSON_SCHEMA_VERSION}')

    raise DatabaseFetchError()


def get_vulnerabilities(pkg: str, spec: str, db: Dict[str, Any]) -> Iterator[Dict[str, Any]]:
    """
    Retrieves vulnerabilities for a package from the database.

    Args:
        pkg (str): The package name.
        spec (str): The specifier set.
        db (Dict[str, Any]): The database.

    Returns:
        Iterator[Dict[str, Any]]: An iterator of vulnerabilities.
    """
    for entry in db['vulnerable_packages'][pkg]:
        for entry_spec in entry["specs"]:
            if entry_spec == spec:
                yield entry


def get_vulnerability_from(
    vuln_id: str,
    cve: Optional[CVE],
    data: Dict[str, Any],
    specifier: str,
    db: Dict[str, Any],
    name: str,
    pkg: Package,
    ignore_vulns: Dict[str, Any],
    affected: SafetyRequirement
) -> Vulnerability:
    """
    Constructs a Vulnerability object from the provided data.

    Args:
        vuln_id (str): The vulnerability ID.
        cve (Optional[CVE]): The CVE object.
        data (Dict[str, Any]): The vulnerability data.
        specifier (str): The specifier set.
        db (Dict[str, Any]): The database.
        name (str): The package name.
        pkg (Package): The Package object.
        ignore_vulns (Dict[str, Any]): The ignored vulnerabilities.
        affected (SafetyRequirement): The affected requirement.

    Returns:
        Vulnerability: The constructed Vulnerability object.
    """
    base_domain = db.get('meta', {}).get('base_domain')
    unpinned_ignored = ignore_vulns.get(vuln_id, {}).get('requirements', None)
    should_ignore = not unpinned_ignored or str(affected.specifier) in unpinned_ignored

    ignored = (ignore_vulns and vuln_id in ignore_vulns and should_ignore and (
            not ignore_vulns[vuln_id]['expires'] or ignore_vulns[vuln_id]['expires'] > datetime.utcnow()))
    more_info_url = f"{base_domain}{data.get('more_info_path', '')}"
    severity = None

    if cve and (cve.cvssv2 or cve.cvssv3):
        severity = Severity(source=cve.name, cvssv2=cve.cvssv2, cvssv3=cve.cvssv3)

    analyzed_requirement = affected
    analyzed_version = next(iter(analyzed_requirement.specifier)).version if is_pinned_requirement(
        analyzed_requirement.specifier) else None

    vulnerable_spec = set()
    vulnerable_spec.add(specifier)

    return Vulnerability(
        vulnerability_id=vuln_id,
        package_name=name,
        pkg=pkg,
        ignored=ignored,
        ignored_reason=ignore_vulns.get(vuln_id, {}).get('reason', None) if ignore_vulns else None,
        ignored_expires=ignore_vulns.get(vuln_id, {}).get('expires', None) if ignore_vulns else None,
        vulnerable_spec=vulnerable_spec,
        all_vulnerable_specs=data.get("specs", []),
        analyzed_version=analyzed_version,
        analyzed_requirement=analyzed_requirement,
        advisory=data.get("advisory"),
        is_transitive=data.get("transitive", False),
        published_date=data.get("published_date"),
        fixed_versions=[ver for ver in data.get("fixed_versions", []) if ver],
        closest_versions_without_known_vulnerabilities=data.get("closest_secure_versions", []),
        resources=data.get("vulnerability_resources"),
        CVE=cve,
        severity=severity,
        affected_versions=data.get("affected_versions", []),
        more_info_url=more_info_url
    )


def get_cve_from(data: Dict[str, Any], db_full: Dict[str, Any]) -> Optional[CVE]:
    """
    Retrieves the CVE object from the provided data.

    Args:
        data (Dict[str, Any]): The vulnerability data.
        db_full (Dict[str, Any]): The full database.

    Returns:
        Optional[CVE]: The CVE object if found, otherwise None.
    """
    try:
        xve_id: str = str(
            next(filter(lambda i: i.get('type', None) in ['cve', 'pve'], data.get('ids', []))).get('id', ''))
    except StopIteration:
        xve_id: str = ''

    if not xve_id:
        return None

    cve_meta = db_full.get("meta", {}).get("severities", {}).get(xve_id, {})
    return CVE(name=xve_id, cvssv2=cve_meta.get("cvssv2", None),
               cvssv3=cve_meta.get("cvssv3", None))


def ignore_vuln_if_needed(
    pkg: Package,
    vuln_id: str,
    cve: Optional[CVE],
    ignore_vulns: Dict[str, Any],
    ignore_severity_rules: Dict[str, Any],
    req: SafetyRequirement
) -> None:
    """
    Determines if a vulnerability should be ignored based on severity rules and updates the ignore_vulns dictionary.

    Args:
        pkg (Package): The package.
        vuln_id (str): The vulnerability ID.
        cve (Optional[CVE]): The CVE object.
        ignore_vulns (Dict[str, Any]): The ignored vulnerabilities.
        ignore_severity_rules (Dict[str, Any]): The severity rules for ignoring vulnerabilities.
        req (SafetyRequirement): The affected requirement.
    """
    if not ignore_severity_rules:
        ignore_severity_rules = {}

    if not isinstance(ignore_vulns, dict):
        return

    severity = None

    if cve:
        if cve.cvssv2 and cve.cvssv2.get("base_score", None):
            severity = cve.cvssv2.get("base_score", None)

        if cve.cvssv3 and cve.cvssv3.get("base_score", None):
            severity = cve.cvssv3.get("base_score", None)

    ignore_severity_below = float(ignore_severity_rules.get('ignore-cvss-severity-below', 0.0))
    ignore_unknown_severity = bool(ignore_severity_rules.get('ignore-cvss-unknown-severity', False))

    if severity:
        if float(severity) < ignore_severity_below:
            reason = 'Ignored by severity rule in policy file, {0} < {1}'.format(float(severity), ignore_severity_below)
            ignore_vulns[vuln_id] = {'reason': reason, 'expires': None}
    elif ignore_unknown_severity:
        reason = 'Unknown CVSS severity, ignored by severity rule in policy file.'
        ignore_vulns[vuln_id] = {'reason': reason, 'expires': None}

    version = next(iter(req.specifier)).version if is_pinned_requirement(req.specifier) else pkg.version

    is_prev_not_ignored: bool = vuln_id not in ignore_vulns
    is_req_not_ignored: bool = 'requirements' in ignore_vulns.get(vuln_id, {}) and str(req.specifier) not in ignore_vulns.get(vuln_id, {}).get('requirements', set())

    if (is_prev_not_ignored or is_req_not_ignored) and is_ignore_unpinned_mode(version):
        reason = IGNORE_UNPINNED_REQ_REASON
        requirements = set()
        requirements.add(str(req.specifier))
        ignore_vulns[vuln_id] = {'reason': reason, 'expires': None, 'requirements': requirements}


def is_vulnerable(vulnerable_spec: SpecifierSet, requirement: SafetyRequirement, package: Package) -> bool:
    """
    Checks if a package version is vulnerable.

    Args:
        vulnerable_spec (SpecifierSet): The specifier set for vulnerable versions.
        requirement (SafetyRequirement): The package requirement.
        package (Package): The package.

    Returns:
        bool: True if the package version is vulnerable, False otherwise.
    """
    if is_pinned_requirement(requirement.specifier):
        try:
            return vulnerable_spec.contains(next(iter(requirement.specifier)).version)
        except Exception:
            # Ugly for now...
            message = f'Version {requirement.specifier} for {package.name} is invalid and is ignored by Safety. Please See PEP 440.'
            if message not in [a['message'] for a in SafetyContext.local_announcements]:
                SafetyContext.local_announcements.append(
                    {'message': message,
                    'type': 'warning', 'local': True})
            return False

    return any(requirement.specifier.filter(vulnerable_spec.filter(package.insecure_versions, prereleases=True),
                                            prereleases=True))


@sync_safety_context
def check(
    *,
    session: requests.Session,
    packages: List[Package] = [],
    db_mirror: Union[Optional[str], bool] = False,
    cached: int = 0,
    ignore_vulns: Optional[Dict[str, Any]] = None,
    ignore_severity_rules: Optional[Dict[str, Any]] = None,
    proxy: Optional[Dict[str, Any]] = None,
    include_ignored: bool = False,
    is_env_scan: bool = True,
    telemetry: bool = True,
    params: Optional[Dict[str, Any]] = None,
    project: Optional[str] = None
) -> tuple:
    """
    Performs a vulnerability check on the provided packages.

    Args:
        session (requests.Session): The requests session.
        packages (List[Package]): The list of packages to check.
        db_mirror (Union[Optional[str], bool]): The database mirror.
        cached (int): The cache validity in seconds.
        ignore_vulns (Optional[Dict[str, Any]]): The ignored vulnerabilities.
        ignore_severity_rules (Optional[Dict[str, Any]]): The severity rules for ignoring vulnerabilities.
        proxy (Optional[Dict[str, Any]]): The proxy settings.
        include_ignored (bool): Whether to include ignored vulnerabilities.
        is_env_scan (bool): Whether it is an environment scan.
        telemetry (bool): Whether to include telemetry data.
        params (Optional[Dict[str, Any]]): Additional parameters.
        project (Optional[str]): The project name.

    Returns:
        tuple: A tuple containing the list of vulnerabilities and the full database.
    """
    SafetyContext().command = 'check'
    db = fetch_database(session, db=db_mirror, cached=cached, telemetry=telemetry)
    db_full = None
    vulnerable_packages = frozenset(db.get('vulnerable_packages', []))
    vulnerabilities = []
    found_pkgs = {}
    requirements = iter([])

    for p in packages:
        requirements = itertools.chain(p.requirements, requirements)
        found_pkgs[canonicalize_name(p.name)] = p

    # Let's report by req, pinned in environment will be ==version
    for req in requirements:
        vuln_per_req = {}
        name = canonicalize_name(req.name)

        pkg = found_pkgs.get(name, None)

        if not pkg.version:
            if not db_full:
                db_full = fetch_database(session, full=True, db=db_mirror, cached=cached,
                                         telemetry=telemetry)
            pkg.refresh_from(db_full)

        if name in vulnerable_packages:
            # we have a candidate here, build the spec set
            for specifier in db['vulnerable_packages'][name]:
                spec_set = SpecifierSet(specifiers=specifier)

                if is_vulnerable(spec_set, req, pkg):
                    if not db_full:
                        db_full = fetch_database(session, full=True, db=db_mirror, cached=cached,
                                                 telemetry=telemetry)
                    if not pkg.latest_version:
                        pkg.refresh_from(db_full)

                    for data in get_vulnerabilities(pkg=name, spec=specifier, db=db_full):
                        try:
                            vuln_id: str = str(next(filter(lambda i: i.get('type', None) == 'pyup', data.get('ids', []))).get('id', ''))
                        except StopIteration:
                            vuln_id: str = ''

                        if vuln_id in vuln_per_req:
                            vuln_per_req[vuln_id].vulnerable_spec.add(specifier)
                            continue

                        cve = get_cve_from(data, db_full)

                        ignore_vuln_if_needed(pkg, vuln_id, cve, ignore_vulns, ignore_severity_rules, req)

                        vulnerability = get_vulnerability_from(vuln_id, cve, data, specifier, db_full, name, pkg,
                                                               ignore_vulns, req)

                        should_add_vuln = not (vulnerability.is_transitive and is_env_scan)

                        if (include_ignored or vulnerability.vulnerability_id not in ignore_vulns) and should_add_vuln:
                            vuln_per_req[vulnerability.vulnerability_id] = vulnerability
                            vulnerabilities.append(vulnerability)

    return vulnerabilities, db_full


def precompute_remediations(
    remediations: Dict[str, Dict[str, Any]],
    packages: Dict[str, Package],
    vulns: List[Vulnerability],
    secure_vulns_by_user: set
) -> None:
    """
    Precomputes the remediations for the given vulnerabilities.

    Args:
        remediations (Dict[str, Dict[str, Any]]): The remediations dictionary.
        packages (Dict[str, Package]): The packages dictionary.
        vulns (List[Vulnerability]): The list of vulnerabilities.
        secure_vulns_by_user (set): The set of vulnerabilities secured by the user.
    """
    for vuln in vulns:

        if vuln.ignored and vuln.ignored_reason != IGNORE_UNPINNED_REQ_REASON:
            secure_vulns_by_user.add(vuln.vulnerability_id)
            continue

        if vuln.package_name in remediations.keys() and str(vuln.analyzed_requirement.specifier) in remediations[vuln.package_name]:
            spec = remediations[vuln.package_name][str(vuln.analyzed_requirement.specifier)]
            spec['vulnerabilities_found'] = spec.get('vulnerabilities_found', 0) + 1
        else:
            version = None
            is_pinned = is_pinned_requirement(vuln.analyzed_requirement.specifier)

            if is_pinned:
                version = next(iter(vuln.analyzed_requirement.specifier)).version

            if not is_pinned and is_ignore_unpinned_mode(version):
                # Let's ignore this requirement
                continue

            vulns_count = 1
            packages[vuln.package_name] = vuln.pkg

            remediations[vuln.package_name][str(vuln.analyzed_requirement.specifier)] = {
                'vulnerabilities_found': vulns_count,
                'version': version,
                'requirement': vuln.analyzed_requirement,
                'more_info_url': vuln.pkg.more_info_url}


def get_closest_ver(
    versions: List[str],
    version: Optional[str],
    spec: SpecifierSet
) -> Dict[str, Optional[Union[str, Version]]]:
    """
    Retrieves the closest versions for the given version and specifier set.

    Args:
        versions (List[str]): The list of versions.
        version (Optional[str]): The current version.
        spec (SpecifierSet): The specifier set.

    Returns:
        Dict[str, Optional[Union[str, Version]]]: The closest versions.
    """
    results: Dict[str, Optional[Union[str, Version]]] = {'upper': None, 'lower': None}

    if (not version and not spec) or not versions:
        return results

    sorted_versions = sorted(versions, key=lambda ver: parse_version(ver), reverse=True)

    if not version:
        sorted_versions = spec.filter(sorted_versions, prereleases=False)

        upper = None
        lower = None

        try:
            sorted_versions = list(sorted_versions)
            upper = sorted_versions[0]
            lower = sorted_versions[-1]
            results['upper'] = upper
            results['lower'] = lower if upper != lower else None
        except IndexError:
            pass

        return results

    current_v = parse_version(version)

    for v in sorted_versions:
        index = parse_version(v)

        if index > current_v:
            results['upper'] = index

        if index < current_v:
            results['lower'] = index
            break

    return results


def compute_sec_ver_for_user(
    package: Package,
    secure_vulns_by_user: set,
    db_full: Dict[str, Any]
) -> List[str]:
    """
    Computes the secure versions for the user.

    Args:
        package (Package): The package.
        secure_vulns_by_user (set): The set of vulnerabilities secured by the user.
        db_full (Dict[str, Any]): The full database.

    Returns:
        List[str]: The list of secure versions.
    """
    versions = package.get_versions(db_full)
    affected_versions = []

    for vuln in db_full.get('vulnerable_packages', {}).get(package.name, []):
        vuln_id: str = str(next(filter(lambda i: i.get('type', None) == 'pyup', vuln.get('ids', []))).get('id', ''))
        if vuln_id and vuln_id not in secure_vulns_by_user:
            affected_versions += vuln.get('affected_versions', [])

    affected_v = set(affected_versions)
    sec_ver_for_user = list(versions.difference(affected_v))

    return sorted(sec_ver_for_user, key=lambda ver: parse_version(ver), reverse=True)


def compute_sec_ver(
    remediations: Dict[str, Dict[str, Any]],
    packages: Dict[str, Package],
    secure_vulns_by_user: set,
    db_full: Dict[str, Any]
) -> None:
    """
    Computes the secure versions and the closest secure version for each remediation.

    Uses the affected_versions of each no ignored vulnerability of the same package, there is only a remediation for each package.

    Args:
        remediations (Dict[str, Dict[str, Any]]): The remediations dictionary.
        packages (Dict[str, Package]): The packages dictionary.
        secure_vulns_by_user (set): The set of vulnerabilities secured by the user.
        db_full (Dict[str, Any]): The full database.
    """
    for pkg_name in remediations.keys():
        pkg: Package = packages.get(pkg_name, None)

        secure_versions = []

        if pkg:
            secure_versions = pkg.secure_versions

        analyzed = set(remediations[pkg_name].keys())

        if not is_using_api_key():
            continue

        for analyzed_requirement in analyzed:
            rem = remediations[pkg_name][analyzed_requirement]
            spec = rem.get('requirement').specifier
            version = rem['version']

            if not secure_vulns_by_user:
                secure_v = sorted(secure_versions, key=lambda ver: parse_version(ver), reverse=True)
            else:
                secure_v = compute_sec_ver_for_user(package=pkg, secure_vulns_by_user=secure_vulns_by_user, db_full=db_full)

            rem['closest_secure_version'] = get_closest_ver(secure_v, version, spec)

            upgrade = rem['closest_secure_version'].get('upper', None)
            downgrade = rem['closest_secure_version'].get('lower', None)
            recommended_version = None

            if upgrade:
                recommended_version = upgrade
            elif downgrade:
                recommended_version = downgrade

            rem['recommended_version'] = recommended_version
            rem['other_recommended_versions'] = [other_v for other_v in secure_v if
                                                                    other_v != str(recommended_version)]

            # Refresh the URL with the recommended version.

            spec = str(rem['requirement'].specifier)
            if is_using_api_key():
                rem['more_info_url'] = \
                    build_remediation_info_url(base_url=rem['more_info_url'], version=version,
                                           spec=spec,
                                           target_version=recommended_version)


def calculate_remediations(
    vulns: List[Vulnerability],
    db_full: Dict[str, Any]
) -> Dict[str, Dict[str, Any]]:
    """
    Calculates the remediations for the given vulnerabilities.

    Args:
        vulns (List[Vulnerability]): The list of vulnerabilities.
        db_full (Dict[str, Any]): The full database.

    Returns:
        Dict[str, Dict[str, Any]]: The calculated remediations.
    """
    remediations = defaultdict(dict)
    package_metadata = {}
    secure_vulns_by_user = set()

    if not db_full:
        return remediations

    precompute_remediations(remediations, package_metadata, vulns, secure_vulns_by_user)
    compute_sec_ver(remediations, package_metadata, secure_vulns_by_user, db_full)

    return remediations


def should_apply_auto_fix(
    from_ver: Optional[Version],
    to_ver: Version,
    allowed_automatic: List[str]
) -> bool:
    """
    Determines if an automatic fix should be applied.

    Args:
        from_ver (Optional[Version]): The current version.
        to_ver (Version): The target version.
        allowed_automatic (List[str]): The allowed automatic update types.

    Returns:
        bool: True if an automatic fix should be applied, False otherwise.
    """
    if not from_ver:
        return False

    if 'major' in allowed_automatic:
        return True

    major_change = to_ver.major - from_ver.major
    minor_change = to_ver.minor - from_ver.minor

    if 'minor' in allowed_automatic:
        if major_change != 0:
            return False

        return True

    if 'patch' in allowed_automatic:
        if major_change != 0 or minor_change != 0:
            return False

        return True

    return False


def get_update_type(from_ver: Optional[Version], to_ver: Version) -> str:
    """
    Determines the update type.

    Args:
        from_ver (Optional[Version]): The current version.
        to_ver (Version): The target version.

    Returns:
        str: The update type.
    """
    if not from_ver or (to_ver.major - from_ver.major) != 0:
        return 'major'

    if (to_ver.minor - from_ver.minor) != 0:
        return 'minor'

    return 'patch'


def process_fixes(
    files: List[str],
    remediations: Dict[str, Dict[str, Any]],
    auto_remediation_limit: List[str],
    output: str,
    no_output: bool = True,
    prompt: bool = False
) -> List[Fix]:
    """
    Processes the fixes for the given files and remediations.

    Args:
        files (List[str]): The list of files.
        remediations (Dict[str, Dict[str, Any]]): The remediations dictionary.
        auto_remediation_limit (List[str]): The automatic remediation limits.
        output (str): The output format.
        no_output (bool): Whether to suppress output.
        prompt (bool): Whether to prompt for confirmation.

    Returns:
        List[Fix]: The list of applied fixes.
    """
    req_remediations = itertools.chain.from_iterable(rem.values() for pkg_name, rem in remediations.items())
    requirements = compute_fixes_per_requirements(files, req_remediations, auto_remediation_limit, prompt=prompt)
    fixes = apply_fixes(requirements, output, no_output, prompt)
    return fixes


def process_fixes_scan(
    file_to_fix: SafetyPolicyFile,
    to_fix_spec: List[SafetyRequirement],
    auto_remediation_limit: List[str],
    output: str,
    no_output: bool = True,
    prompt: bool = False
) -> List[Fix]:
    """
    Processes the fixes for the given file and specifications in scan mode.

    Args:
        file_to_fix (SafetyPolicyFile): The file to fix.
        to_fix_spec (List[SafetyRequirement]): The specifications to fix.
        auto_remediation_limit (List[str]): The automatic remediation limits.
        output (str): The output format.
        no_output (bool): Whether to suppress output.
        prompt (bool): Whether to prompt for confirmation.

    Returns:
        List[Fix]: The list of applied fixes.
    """
    to_fix_remediations =  []

    def get_remmediation_from(spec):
        upper = None
        lower = None
        recommended = None

        try:
            upper = Version(spec.remediation.closest_secure.upper) if spec.remediation.closest_secure.upper else None
        except Exception as e:
            LOG.error(f'Error getting upper remediation version, ignoring', exc_info=True)

        try:
            lower = Version(spec.remediation.closest_secure.lower) if spec.remediation.closest_secure.lower else None
        except Exception as e:
            LOG.error(f'Error getting lower remediation version, ignoring', exc_info=True)

        try:
            recommended = Version(spec.remediation.recommended)
        except Exception as e:
            LOG.error(f'Error getting recommended version for remediation, ignoring', exc_info=True)

        return {
            "vulnerabilities_found": spec.remediation.vulnerabilities_found,
            "version": next(iter(spec.specifier)).version if spec.is_pinned() else None,
            "requirement": spec,
            "more_info_url": spec.remediation.more_info_url,
            "closest_secure_version": {
                'upper': upper,
                'lower': lower
                },
            "recommended_version": recommended,
            "other_recommended_versions": spec.remediation.other_recommended
        }

    req_remediations = iter(get_remmediation_from(spec) for spec in to_fix_spec)
    SUPPORTED_FILE_TYPES = [FileType.REQUIREMENTS_TXT]

    if file_to_fix.file_type in SUPPORTED_FILE_TYPES:
        files = (open(file_to_fix.location),)
        requirements = compute_fixes_per_requirements(files, req_remediations, auto_remediation_limit, prompt=prompt)
    else:
        requirements = {
            'files': {str(file_to_fix.location): {'content': None, 'fixes': {'TO_SKIP': [], 'TO_APPLY': [], 'TO_CONFIRM': []}, 'supported': False, 'filename': file_to_fix.location.name}},
            'dependencies': defaultdict(dict),
        }

    fixes = apply_fixes(requirements, output, no_output, prompt, scan_flow=True, auto_remediation_limit=auto_remediation_limit)

    return fixes


def compute_fixes_per_requirements(
    files: List[str],
    req_remediations: Iterator[Dict[str, Any]],
    auto_remediation_limit: List[str],
    prompt: bool = False
) -> Dict[str, Any]:
    """
    Computes the fixes per requirements.

    Args:
        files (List[str]): The list of files.
        req_remediations (Iterator[Dict[str, Any]]): The remediations iterator.
        auto_remediation_limit (List[str]): The automatic remediation limits.
        prompt (bool): Whether to prompt for confirmation.

    Returns:
        Dict[str, Any]: The computed requirements with fixes.
    """
    requirements_files = get_requirements_content(files)

    from dparse.parser import parse, filetypes
    from packaging.version import Version, InvalidVersion

    requirements = {
        'files': {},
        'dependencies': defaultdict(dict),
    }

    for name, contents in requirements_files.items():
        dependency_file = parse(contents, path=name, file_type=filetypes.requirements_txt, resolve=True)
        dependency_files = dependency_file.resolved_files + [dependency_file]
        empty_spec = SpecifierSet('')
        default_spec = SpecifierSet('>=0')

        # Support recursive requirements in the multiple requirement files provided
        for resolved_f in dependency_files:
            if not resolved_f or isinstance(resolved_f, str):
                continue
            file = {'content': resolved_f.content, 'fixes': {'TO_SKIP': [], 'TO_APPLY': [], 'TO_CONFIRM': []}}
            requirements['files'][resolved_f.path] = file

            for d in resolved_f.dependencies:
                if d.specs == empty_spec:
                    d.specs = default_spec

                requirements['dependencies'][d.name][str(d.specs)] = (d, resolved_f.path)

    for remediation in req_remediations:
        req: SafetyRequirement = remediation.get('requirement')
        pkg: str = req.name

        dry_fix = Fix(package=pkg, more_info_url=remediation.get('more_info_url', ''),
                      previous_spec=str(req.specifier),
                      other_options=remediation.get('other_recommended_versions', []))
        from_ver: Optional[str] = remediation.get('version', None)

        if pkg not in requirements['dependencies'] or dry_fix.previous_spec not in requirements['dependencies'][pkg]:
            # Let's attach it to the first file scanned.
            file = next(iter(requirements['files']))
            # Let's use the no parsed version.
            dry_fix.previous_version = from_ver
            dry_fix.status = 'AUTOMATICALLY_SKIPPED_NOT_FOUND_IN_FILE'
            dry_fix.applied_at = file
            requirements['files'][file]['fixes']['TO_SKIP'].append(dry_fix)
            continue

        dependency, name = requirements['dependencies'][pkg][dry_fix.previous_spec]
        dry_fix.applied_at = name

        fixes = requirements['files'][name]['fixes']

        to_ver: Version = remediation['recommended_version']

        try:
            from_ver = parse_version(from_ver)
        except (InvalidVersion, TypeError):

            if not dry_fix.previous_spec:
                dry_fix.status = 'AUTOMATICALLY_SKIPPED_INVALID_VERSION'
                fixes['TO_SKIP'].append(dry_fix)
                continue

        dry_fix.previous_version = str(from_ver) if from_ver else from_ver

        if remediation['recommended_version'] is None:
            dry_fix.status = 'AUTOMATICALLY_SKIPPED_NO_RECOMMENDED_VERSION'
            fixes['TO_SKIP'].append(dry_fix)
            continue

        dry_fix.updated_version = str(to_ver)

        is_fixed = from_ver == to_ver

        if is_fixed:
            dry_fix.status = 'AUTOMATICALLY_SKIPPED_ALREADY_FIXED'
            fixes['TO_SKIP'].append(dry_fix)
            continue

        update_type = get_update_type(from_ver, to_ver)
        dry_fix.update_type = update_type
        dry_fix.dependency = dependency

        auto_fix = should_apply_auto_fix(from_ver, to_ver, auto_remediation_limit)

        TARGET = 'TO_APPLY'

        if auto_fix:
            dry_fix.status = 'PENDING_TO_APPLY'
            dry_fix.fix_type = 'AUTOMATIC'
        elif prompt:
            TARGET = 'TO_CONFIRM'
            dry_fix.status = 'PENDING_TO_CONFIRM'
            dry_fix.fix_type = 'MANUAL'
        else:
            TARGET = 'TO_SKIP'
            dry_fix.status = 'AUTOMATICALLY_SKIPPED_UNABLE_TO_CONFIRM'

        fixes[TARGET].append(dry_fix)

    return requirements


def apply_fixes(
    requirements: Dict[str, Any],
    out_type: str,
    no_output: bool,
    prompt: bool,
    scan_flow: bool = False,
    auto_remediation_limit: List[str] = None
) -> List[Fix]:
    """
    Applies the fixes to the requirements.

    Args:
        requirements (Dict[str, Any]): The requirements with fixes.
        out_type (str): The output format.
        no_output (bool): Whether to suppress output.
        prompt (bool): Whether to prompt for confirmation.
        scan_flow (bool): Whether it is in scan flow mode.
        auto_remediation_limit (List[str]): The automatic remediation limits.

    Returns:
        List[Fix]: The list of applied fixes.
    """

    from dparse.updater import RequirementsTXTUpdater

    skip = []
    apply = []
    confirm = []

    brief = []

    if not no_output:
        style_kwargs = {}

        if not scan_flow:
            brief.append(('', {}))
            brief.append((f"Safety fix running", style_kwargs))
        print_service(brief, out_type)

    for name, data in requirements['files'].items():
        output = [('', {}),
                  (f"Analyzing {name}... [{get_fix_opt_used_msg(auto_remediation_limit)} limit]", {'styling': {'bold': True}, 'start_line_decorator': '->', 'indent': ' '})]

        r_skip = data['fixes']['TO_SKIP']
        r_apply = data['fixes']['TO_APPLY']
        r_confirm = data['fixes']['TO_CONFIRM']

        if data.get('supported', True):
            new_content = data['content']

            updated: bool = False

            for f in r_apply:
                new_content = RequirementsTXTUpdater.update(content=new_content, version=f.updated_version,
                                                            dependency=f.dependency, hashes=get_hashes(f.dependency))
                f.status = 'APPLIED'
                updated = True
                output.append(('', {}))
                output.append((f'- {get_applied_msg(f, mode="auto")}', {}))

            for f in r_skip:
                output.append(('', {}))
                output.append((f'- {get_skipped_msg(f)}', {}))

            if not no_output:
                print_service(output, out_type)

            if prompt and not no_output:
                for f in r_confirm:
                    options = [f"({index}) =={option}" for index, option in enumerate(f.other_options)]
                    input_hint = f'Enter “y” to update to {f.package}=={f.updated_version}, “n” to skip this package upgrade'

                    if len(options) > 0:
                        input_hint += f', or enter the index from these secure versions to upgrade to that version: {", ".join(options)}'

                    print_service([('', {})], out_type)
                    confirmed: str = prompt_service(
                        (f'- {f.package}{f.previous_spec} requires at least a {f.update_type} version update. Do you want to update {f.package} from {f.previous_spec} to =={f.updated_version}, which is the closest secure version? {input_hint}', {}),
                        out_type
                    ).lower()

                    try:
                        index: int = int(confirmed)
                        if index <= len(f.other_options):
                            confirmed = 'y'
                    except ValueError:
                        index = -1

                    if confirmed == 'y' or index > -1:
                        f.status = 'APPLIED'
                        updated = True

                        if index > -1:
                            f.updated_version = f.other_options[index]

                        new_content = RequirementsTXTUpdater.update(content=new_content, version=f.updated_version,
                                                                    dependency=f.dependency,
                                                                    hashes=get_hashes(f.dependency))
                        output.append((get_applied_msg(f, mode="manual"), {'indent': ' ' * 5}))
                    else:
                        f.status = 'MANUALLY_SKIPPED'
                        output.append((get_skipped_msg(f), {'indent': ' ' * 5}))

                    if not no_output:
                        print_service(output, out_type)

            if updated:
                output.append(('', {}))
                output.append((f"Updating {name}...", {}))
                with open(name, mode="w") as r_file:
                    r_file.write(new_content)
                output.append((f"Changes applied to {name}.", {}))
                count = len(r_apply) + len([1 for fix in r_confirm if fix.status == 'APPLIED'])
                output.append((f"{count} package {pluralize('version', count)} {pluralize('has', count)} been updated to secure versions in {Path(name).name}", {}))
                output.append(("Always check for breaking changes after updating packages.", {}))
            else:
                output.append((f"No fixes to be made in {name}.", {}))
                output.append(('', {}))
        else:
            not_supported_filename = data.get('filename', name)
            output.append(
                (f"{not_supported_filename} updates not supported: Please update these dependencies using your package manager.",
                 {'start_line_decorator': ' -', 'indent': ' '}))
            output.append(('', {}))

        if not no_output:
            print_service(output, out_type)

        skip.extend(r_skip)
        apply.extend(r_apply)
        confirm.extend(r_confirm)

    # The scan flow will handle the header and divider, because the scan flow can be called multiple times.
    if not no_output and not scan_flow:
        divider = f'{"=" * 78}' if out_type == 'text' else f'{"=" * (get_terminal_size().columns - 2)}'
        format_text = {'start_line_decorator': '+', 'end_line_decorator': '+', 'indent': ''}
        print_service([(divider, {})], out_type, format_text=format_text)

    return skip + apply + confirm


def find_vulnerabilities_fixed(
    vulnerabilities: Dict[str, Any],
    fixes: List[Fix]
) -> List[Vulnerability]:
    """
    Finds the vulnerabilities that have been fixed.

    Args:
        vulnerabilities (Dict[str, Any]): The dictionary of vulnerabilities.
        fixes (List[Fix]): The list of applied fixes.

    Returns:
        List[Vulnerability]: The list of fixed vulnerabilities.
    """
    fixed_specs = set(fix.previous_spec for fix in fixes)

    if not fixed_specs:
        return []

    return [vulnerability for vulnerability in vulnerabilities if
            str(vulnerability['analyzed_requirement'].specifier) in fixed_specs]


@sync_safety_context
def review(
    *,
    report: Optional[Dict[str, Any]] = None,
    params: Optional[Dict[str, Any]] = None
) -> tuple:
    """
    Reviews the report and returns the vulnerabilities and remediations.

    Args:
        report (Optional[Dict[str, Any]]): The report.
        params (Optional[Dict[str, Any]]): Additional parameters.

    Returns:
        tuple: A tuple containing the list of vulnerabilities, the remediations, and the found packages.
    """
    SafetyContext().command = 'review'
    vulnerable = []
    vulnerabilities = report.get('vulnerabilities', []) + report.get('ignored_vulnerabilities', [])
    remediations = defaultdict(dict)

    for key, pkg_rem in report.get('remediations', {}).items():
        analyzed = set(pkg_rem['requirements'].keys())

        for req in analyzed:
            req_rem = pkg_rem['requirements'][req]
            recommended = req_rem.get('recommended_version', None)
            secure_v = req_rem.get('other_recommended_versions', [])

            remediations[key][req] = {'vulnerabilities_found': req_rem.get('vulnerabilities_found', 0),
                                      'version': req_rem.get('version'),
                                      'requirement': SafetyRequirement(req_rem['requirement']['raw']),
                                      'other_recommended_versions': secure_v,
                                      'recommended_version': parse_version(recommended) if recommended else None,
                                      # minor isn't supported in review
                                      'more_info_url': req_rem.get('more_info_url')}

    packages = report.get('scanned_packages', [])
    pkgs = {}

    for name, values in packages.items():
        requirements = [SafetyRequirement(r['raw']) for r in values.get('requirements', [])]
        values.update({'requirements': requirements})
        pkgs[name] = Package(**values)

    ctx = SafetyContext()
    found_packages = list(pkgs.values())
    ctx.packages = found_packages
    ctx.review = report.get('report_meta', [])
    ctx.key = ctx.review.get('api_key', False)
    cvssv2 = None
    cvssv3 = None

    for vuln in vulnerabilities:
        vuln['pkg'] = pkgs.get(vuln.get('package_name', None))
        XVE_ID = vuln.get('CVE', None)  # Trying to get first the CVE ID

        severity = vuln.get('severity', None)
        if severity and severity.get('source', False):
            cvssv2 = severity.get('cvssv2', None)
            cvssv3 = severity.get('cvssv3', None)
            # Trying to get the PVE ID if it exists, otherwise it will be the same CVE ID of above
            XVE_ID = severity.get('source', False)
            vuln['severity'] = Severity(source=XVE_ID, cvssv2=cvssv2, cvssv3=cvssv3)
        else:
            vuln['severity'] = None

        ignored_expires = vuln.get('ignored_expires', None)

        if ignored_expires:
            vuln['ignored_expires'] = validate_expiration_date(ignored_expires)

        vuln['CVE'] = CVE(name=XVE_ID, cvssv2=cvssv2, cvssv3=cvssv3) if XVE_ID else None
        vuln['analyzed_requirement'] = SafetyRequirement(vuln['analyzed_requirement']['raw'])

        vulnerable.append(Vulnerability(**vuln))

    return vulnerable, remediations, found_packages


@sync_safety_context
def get_licenses(
    *,
    session: requests.Session,
    db_mirror: Union[Optional[str], bool] = False,
    cached: int = 0,
    telemetry: bool = True
) -> Dict[str, Any]:
    """
    Retrieves the licenses from the database.

    Args:
        session (requests.Session): The requests session.
        db_mirror (Union[Optional[str], bool]): The database mirror.
        cached (int): The cache validity in seconds.
        telemetry (bool): Whether to include telemetry data.

    Returns:
        Dict[str, Any]: The licenses dictionary.
    """
    if db_mirror:
        mirrors = [db_mirror]
    else:
        mirrors = API_MIRRORS

    db_name = "licenses.json"

    for mirror in mirrors:
        # mirror can either be a local path or a URL
        if is_a_remote_mirror(mirror):
            licenses = fetch_database_url(session, mirror, db_name=db_name, cached=cached,
                                          telemetry=telemetry)
        else:
            licenses = fetch_database_file(mirror, db_name=db_name, ecosystem=None)
        if licenses:
            return licenses
    raise DatabaseFetchError()


def add_local_notifications(
    packages: List[Package],
    ignore_unpinned_requirements: Optional[bool]
) -> List[Dict[str, str]]:
    """
    Adds local notifications for unpinned packages.

    Args:
        packages (List[Package]): The list of packages.
        ignore_unpinned_requirements (Optional[bool]): Whether to ignore unpinned requirements.

    Returns:
        List[Dict[str, str]]: The list of notifications.
    """
    announcements = []
    unpinned_packages: List[str] = [f"{pkg.name}" for pkg in packages if pkg.has_unpinned_req()]

    if unpinned_packages and ignore_unpinned_requirements is not False:
        found = len(unpinned_packages)
        and_msg = ''

        if found >= 2:
            last = unpinned_packages.pop()
            and_msg = f' and {last}'

        pkgs: str = f"{', '.join(unpinned_packages)}{and_msg} {'are' if found > 1 else 'is'}"
        doc_msg: str = get_specifier_range_info(style=False, pin_hint=True)

        if ignore_unpinned_requirements is None:
            msg = f'Warning: {pkgs} unpinned. Safety by default does not ' \
                  f'report on potential vulnerabilities in unpinned packages. {doc_msg}'
        else:
            msg = f'Warning: {pkgs} unpinned and potential vulnerabilities are ' \
                  f'being ignored given `ignore-unpinned-requirements` is True in your config. {doc_msg}'

        announcements.append({'message': msg, 'type': 'warning', 'local': True})

    announcements.extend(SafetyContext().local_announcements)

    return announcements


def get_announcements(
    session: requests.Session,
    telemetry: bool = True,
    with_telemetry: Any = None
) -> List[Dict[str, str]]:
    """
    Retrieves announcements from the server.

    Args:
        session (requests.Session): The requests session.
        telemetry (bool): Whether to include telemetry data.
        with_telemetry (Optional[Dict[str, Any]]): The telemetry data.

    Returns:
        List[Dict[str, str]]: The list of announcements.
    """
    LOG.info('Getting announcements')

    announcements = []

    url = f"{DATA_API_BASE_URL}announcements/"
    method = 'post'
    telemetry_data = with_telemetry if with_telemetry else build_telemetry_data(telemetry=telemetry)
    data = asdict(telemetry_data)
    request_kwargs = {'timeout': 3}
    data_keyword = 'json'

    source = os.environ.get('SAFETY_ANNOUNCEMENTS_URL', None)

    if source:
        LOG.debug(f'Getting the announcement from a different source: {source}')
        url = source
        method = 'get'
        data = {
            'telemetry': json.dumps(data)}
        data_keyword = 'params'

    request_kwargs[data_keyword] = data
    request_kwargs['url'] = url

    LOG.debug(f'Telemetry data sent: {data}')

    try:
        request_func = getattr(session, method)
        r = request_func(**request_kwargs)
        LOG.debug(r.text)
    except Exception as e:
        LOG.info('Unexpected but HANDLED Exception happened getting the announcements: %s', e)
        return announcements

    if r.status_code == 200:
        try:
            announcements = r.json()
            if 'announcements' in announcements.keys():
                announcements = announcements.get('announcements', [])
            else:
                LOG.info('There is not announcements key in the JSON response, is this a wrong structure?')
                announcements = []

        except json.JSONDecodeError as e:
            LOG.info('Unexpected but HANDLED Exception happened decoding the announcement response: %s', e)

    LOG.info('Announcements fetched')

    return announcements



def get_packages(files: Optional[List[str]] = None, stdin: bool = False) -> List[Package]:
    """
    Retrieves the packages from the given files or standard input.

    Args:
        files (Optional[List[str]]): The list of files.
        stdin (bool): Whether to read from standard input.

    Returns:
        List[Package]: The list of packages.
    """
    if files:
        return list(itertools.chain.from_iterable(read_requirements(f, resolve=True) for f in files))

    if stdin:
        return list(read_requirements(sys.stdin))

    # TODO: Migrate away from pkg_resources and use importlib
    import pkg_resources

    def allowed_version(pkg: str, version: str) -> bool:
        try:
            parse_version(version)
        except Exception:
            SafetyContext.local_announcements.append(
                {'message': f'Version {version} for {pkg} is invalid and is ignored by Safety. Please See PEP 440.',
                 'type': 'warning', 'local': True})
            return False

        return True

    w_set = pkg_resources.working_set

    SafetyContext().scanned_full_path.extend(w_set.entry_keys.keys())

    return [
        Package(name=d.key, version=d.version,
                requirements=[SafetyRequirement(f'{d.key}=={d.version}', found=d.location)],
                found=d.location, insecure_versions=[],
                secure_versions=[], latest_version=None, latest_version_without_known_vulnerabilities=None,
                more_info_url=None) for d in
        w_set
        if d.key not in {"python", "wsgiref", "argparse"} and allowed_version(d.key, d.version)
    ]


def read_vulnerabilities(fh: Any) -> Dict[str, Any]:
    """
    Reads vulnerabilities from a file handle.

    Args:
        fh (Any): The file handle.

    Returns:
        Dict[str, Any]: The vulnerabilities data.
    """
    try:
        data = json.load(fh)
    except json.JSONDecodeError as e:
        raise MalformedDatabase(reason=e, fetched_from=fh.name)
    except TypeError as e:
        raise MalformedDatabase(reason=e, fetched_from=fh.name)

    return data


def get_server_policies(
    session: requests.Session,
    policy_file: SafetyPolicyFile,
    proxy_dictionary: Dict[str, str]
) -> tuple:
    """
    Retrieves the server policies.

    Args:
        session (requests.Session): The requests session.
        policy_file (SafetyPolicyFile): The policy file.
        proxy_dictionary (Dict[str, str]): The proxy dictionary.

    Returns:
        tuple: A tuple containing the policy file and the audit and monitor flag.
    """
    if session.api_key:
        server_policies = fetch_policy(session)
        server_audit_and_monitor = server_policies["audit_and_monitor"]
        server_safety_policy = server_policies["safety_policy"]
    else:
        server_audit_and_monitor = False
        server_safety_policy = ""

    if server_safety_policy and policy_file:
        click.secho(
            "Warning: both a local policy file '{policy_filename}' and a server sent policy are present. "
            "Continuing with the local policy file.".format(policy_filename=policy_file['filename']),
            fg="yellow",
            file=sys.stderr
        )
    elif server_safety_policy:
        with tempfile.NamedTemporaryFile(prefix='server-safety-policy-') as tmp:
            tmp.write(server_safety_policy.encode('utf-8'))
            tmp.seek(0)

            policy_file = SafetyPolicyFile().convert(tmp.name, param=None, ctx=None)
            LOG.info('Using server side policy file')

    return policy_file, server_audit_and_monitor


def save_report(
    path: str,
    default_name: str,
    report: str
) -> None:
    """
    Saves the report to a file.

    Args:
        path (str): The path to save the report.
        default_name (str): The default name of the report file.
        report (str): The report content.
    """
    if path:
        save_at = path

        if os.path.isdir(path):
            save_at = os.path.join(path, default_name)

        with open(save_at, 'w+') as report_file:
            report_file.write(report)



import logging
import os
from pathlib import Path
import shutil
import sys
from typing import Any, Dict, Optional
import tomlkit

from rich.console import Console
from safety.console import main_console
from safety.tool.constants import (
    ORGANIZATION_REPOSITORY_URL,
    PUBLIC_REPOSITORY_URL,
    PROJECT_REPOSITORY_URL,
)

if sys.version_info >= (3, 11):
    import tomllib
else:
    import tomli as tomllib

logger = logging.getLogger(__name__)


def backup_file(path: Path) -> None:
    """
    Create backup of file if it exists
    """
    if path.exists():
        backup_path = path.with_name(f"{path.name}.backup")
        shutil.copy2(path, backup_path)


class Uv:
    @classmethod
    def is_installed(cls) -> bool:
        """
        Checks if the UV program is installed

        Returns:
            True if UV is installed on system, or false otherwise
        """
        return shutil.which("uv") is not None

    @classmethod
    def is_uv_project_file(cls, file: Path) -> bool:
        try:
            cfg = tomllib.loads(file.read_text())
            return (
                cfg.get("tool", {}).get("uv") is not None
                or (file.parent / "uv.lock").exists()
            )
        except (IOError, ValueError):
            return False

    @classmethod
    def configure_pyproject(
        cls,
        file: Path,
        org_slug: Optional[str],
        project_id: Optional[str] = None,
        console: Console = main_console,
    ) -> Optional[Path]:
        """
        Configures index url for specified pyproject.toml file.

        Args:
            file (Path): Path to pyproject.toml file.
            org_slug (Optional[str]): Organization slug.
            project_id (Optional[str]): Project ID.
            console (Console): Console instance.
        """
        if not cls.is_installed():
            logger.error("UV is not installed.")
            return None

        repository_url = (
            PROJECT_REPOSITORY_URL.format(org_slug, project_id)
            if project_id and org_slug
            else (
                ORGANIZATION_REPOSITORY_URL.format(org_slug)
                if org_slug
                else PUBLIC_REPOSITORY_URL
            )
        )
        try:
            content = file.read_text()
            doc: Dict[str, Any] = tomlkit.loads(content)

            if "tool" not in doc:
                doc["tool"] = tomlkit.table()
            if "uv" not in doc["tool"]:  # type: ignore
                doc["tool"]["uv"] = tomlkit.table()  # type: ignore
            if "index" not in doc["tool"]["uv"]:  # type: ignore
                doc["tool"]["uv"]["index"] = tomlkit.aot()  # type: ignore

            index_container = doc["tool"]["uv"]  # type: ignore
            cls.filter_out_safety_index(index_container)

            safety_index = {
                "name": "safety",
                "url": repository_url,
                # In UV default:
                # True = lowest priority
                # False = highest priority
                "default": False,
            }
            non_safety_indexes = (
                doc.get("tool", {}).get("uv", {}).get("index", tomlkit.aot())
            )

            # Add safety index as first priority
            index_container["index"] = tomlkit.aot()  # type: ignore
            index_container["index"].append(safety_index)  # type: ignore
            index_container["index"].extend(non_safety_indexes)  # type: ignore

            # Write back to file
            file.write_text(tomlkit.dumps(doc))
            return file

        except (IOError, ValueError, Exception) as e:
            logger.error(f"Failed to configure {file} file: {e}")

        return None

    @classmethod
    def get_user_config_path(cls) -> Path:
        """
        Returns the path to the user config file for UV.

        This logic is based on the uv documentation:
        https://docs.astral.sh/uv/configuration/files/

        "uv will also discover user-level configuration at
        ~/.config/uv/uv.toml (or $XDG_CONFIG_HOME/uv/uv.toml) on macOS and Linux,
        or %APPDATA%\\uv\\uv.toml on Windows; ..."

        Returns:
            Path: The path to the user config file.
        """
        if sys.platform == "win32":
            return Path(os.environ.get("APPDATA", ""), "uv", "uv.toml")
        else:
            xdg_config_home = os.environ.get("XDG_CONFIG_HOME")
            if xdg_config_home:
                return Path(xdg_config_home, "uv", "uv.toml")
            else:
                return Path(Path.home(), ".config", "uv", "uv.toml")

    @classmethod
    def filter_out_safety_index(cls, index_container: Any):
        if "index" not in index_container:
            return

        indexes = list(index_container["index"])
        index_container["index"] = tomlkit.aot()

        for index in indexes:
            index_url = index.get("url", "")

            if ".safetycli.com" in index_url:
                continue

            index_container["index"].append(index)

    @classmethod
    def configure_system(
        cls, org_slug: Optional[str], console: Console = main_console
    ) -> Optional[Path]:
        """
        Configures UV system to use to Safety index url.
        """
        try:
            repository_url = (
                ORGANIZATION_REPOSITORY_URL.format(org_slug)
                if org_slug
                else PUBLIC_REPOSITORY_URL
            )

            user_config_path = cls.get_user_config_path()

            if not user_config_path.exists():
                user_config_path.parent.mkdir(parents=True, exist_ok=True)
                content = ""
            else:
                backup_file(user_config_path)
                content = user_config_path.read_text()

            doc = tomlkit.loads(content)
            if "index" not in doc:
                doc["index"] = tomlkit.aot()
            cls.filter_out_safety_index(index_container=doc)

            safety_index = tomlkit.aot()
            safety_index.append(
                {
                    "name": "safety",
                    "url": repository_url,
                    # In UV default:
                    # True = lowest priority
                    # False = highest priority
                    "default": False,
                }
            )

            non_safety_indexes = doc.get("index", tomlkit.aot())

            # Add safety index as first priority
            doc["index"] = tomlkit.aot()
            doc.append("index", safety_index)
            doc.append("index", non_safety_indexes)

            user_config_path.write_text(tomlkit.dumps(doc))
            return user_config_path

        except Exception as e:
            logger.error(f"Failed to configure UV system: {e}")
            return None

    @classmethod
    def reset_system(cls, console: Console = main_console):
        try:
            user_config_path = cls.get_user_config_path()
            if user_config_path.exists():
                backup_file(user_config_path)
                content = user_config_path.read_text()
                doc = tomlkit.loads(content)
                cls.filter_out_safety_index(index_container=doc)
                user_config_path.write_text(tomlkit.dumps(doc))
        except Exception as e:
            msg = "Failed to reset UV global settings"
            logger.error(f"{msg}: {e}")
